#!/usr/bin/env  python

from argparse import ArgumentParser

from pyxtal import print_logo
from pyxtal.symmetry import Group

if __name__ == "__main__":
    # -------------------------------- Options -------------------------
    parser = ArgumentParser()
    parser.add_argument(
        "-s",
        "--symmetry",
        dest="sg",
        type=str,
        help="desired symmetry, number or string, e.g., 36, Pbca, Ih. if None, show all list of available groups",
    )
    parser.add_argument(
        "-d",
        "--dimension",
        dest="dimension",
        default=3,
        type=int,
        help="desired dimension: (3, 2, 1, 0): default 3",
    )

    print_logo()
    options = parser.parse_args()

    dimension = options.dimension

    if options.sg is not None:
        sg = options.sg
        if sg.isnumeric():
            sg = int(sg)
        print(Group(sg, dimension))
    else:
        Group.list_groups(dimension)



import numpy as np
from scipy.spatial.distance import cdist, pdist

"""
LJ energy and force functions
"""


def LJ(pos, dim, mu=0.1, shift=False):
    """
    Calculate the total energy
    Args:
    pos: 1D array with N*dim numbers representing the atomic positions
    dim: dimension of the hyper/normal space
    output
    E: the total energy with punishing function
    """
    N_atom = int(len(pos) / dim)
    pos = np.reshape(pos, (N_atom, dim))

    distance = pdist(pos)
    r6 = np.power(distance, 6)
    r12 = np.multiply(r6, r6)
    Eng = np.sum(4 * (1 / r12 - 1 / r6))

    if dim > 3:
        norm = 0
        for i in range(3, dim):
            diff = pos[:, i] - np.mean(pos[:, i]) if shift else pos[:, i]
            norm += np.sum(np.power(diff, 2))
        Eng += 0.5 * mu * norm
    return Eng


def LJ_force(pos, dim, mu=0.1, shift=False):
    N_atom = int(len(pos) / dim)
    pos = np.reshape(pos, [N_atom, dim])
    force = np.zeros([N_atom, dim])
    for i, pos0 in enumerate(pos):
        pos1 = pos.copy()
        pos1 = np.delete(pos1, i, 0)
        distance = cdist([pos0], pos1)
        r = pos1 - pos0
        r2 = np.power(distance, 2)
        r6 = np.power(r2, 3)
        r12 = np.power(r6, 2)
        force[i] = np.dot((48 / r12 - 24 / r6) / r2, r)
        # force from the punish function mu*sum([x-mean(x)]^2)
        if dim > 3:
            for j in range(3, dim):
                if shift:
                    force[i, j] += mu * (pos[i, j] - np.mean(pos[:, j]))
                else:
                    force[i, j] += mu * pos[i, j]
    return force.flatten()



# Standard Libraries
from __future__ import annotations

import numpy as np
from numpy.random import Generator

from pyxtal.constants import deg, ltype_keywords, rad

# PyXtal imports
from pyxtal.msg import VolumeError
from pyxtal.operations import angle, create_matrix


class Lattice:
    """
    Class for storing and generating crystal lattices. Allows for
    specification of constraint values. Lattice types include triclinic,
    monoclinic, orthorhombic, tetragonal, trigonal, hexagonal, cubic,
    spherical, and ellipsoidal. The last two are used for generating point
    group structures, and do not actually represent a parallelepiped lattice.

    Args:
        ltype: a string representing the type of lattice (from the above list)
        volume: the volume, in Angstroms cubed, of the lattice
        matrix: matrix in 3*3 form
        PBC: A periodic boundary condition list, where 1 is periodic,
            Ex: [1, 1, 1] -> 3d periodicity, [0, 0, 1] -> periodic at z axis
        kwargs: various values which may be defined. If none are defined,
            random ones will be generated. Values will be passed to
            generate_lattice. Options include:
            area: The cross-sectional area (in Ang^2). Only for 1D crystals
            thickness: The cell's thickness (in Angstroms) for 2D crystals
            unique_axis: The unique axis for certain symmetry (and especially
                layer) groups. Because the symmetry operations are not also
                transformed, you should use the default values for random
                crystal generation
            random: If False, keeps the stored values for the lattice geometry
                even upon applying reset_matrix. To alter the matrix, use
                set_matrix() or set_para
            'unique_axis': the axis ('a', 'b', or 'c') which is not symmetrically
                equivalent to the other two
            'min_l': the smallest allowed cell vector. The smallest vector must
                be larger than this.
            'mid_l': the second smallest allowed cell vector. The second
                smallest vector must be larger than this.
            'max_l': the third smallest allowed cell vector. The largest cell
                vector must be larger than this.
            'allow_volume_reset': a bool stating whether or not the volume
                should be reset during each crystal generation attempt
    """

    def __init__(
        self, ltype, volume=None, matrix=None, PBC=None, random_state: int | None | Generator = None, **kwargs
    ):
        # Set required parameters
        if PBC is None:
            PBC = [1, 1, 1]

        if ltype in ltype_keywords:
            self.ltype = ltype.lower()
        elif ltype is None:
            self.ltype = "triclinic"
        else:
            msg = "Invalid lattice type: " + ltype
            raise ValueError(msg)

        self.volume = float(volume)
        self.PBC = PBC
        self.dim = sum(PBC)
        self.kwargs = {}
        self.random = True

        if isinstance(random_state, Generator):
            self.random_state = random_state.spawn(1)[0]
        else:
            self.random_state = np.random.default_rng(random_state)

        # Set optional values
        self.allow_volume_reset = True
        for key, value in kwargs.items():
            if key in [
                "area",
                "thickness",
                "unique_axis",
                "random",
                "min_l",
                "mid_l",
                "max_l",
                "min_special",  # min special for mole
            ]:
                setattr(self, key, value)
                self.kwargs[key] = value
                if key == "allow_volume_reset" and value is False:
                    self.allow_volume_reset = False

        if not hasattr(self, "unique_axis"):
            self.unique_axis = "c"

        # Set stress normalization info
        if self.ltype == "triclinic":
            norm_matrix = np.ones([3, 3])

        elif self.ltype == "monoclinic":
            if self.PBC == [1, 1, 1]:
                norm_matrix = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 1]])
            else:
                if self.unique_axis == "a":
                    norm_matrix = np.array([[1, 0, 0], [0, 1, 0], [0, 1, 1]])
                elif self.unique_axis == "b":
                    norm_matrix = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 1]])
                elif self.unique_axis == "c":
                    norm_matrix = np.array([[1, 0, 0], [1, 1, 0], [0, 0, 1]])

        elif self.ltype in [
            "orthorhombic",
            "tetragonal",
            "trigonal",
            "hexagonal",
            "cubic",
        ]:
            norm_matrix = np.eye(3)

        elif self.ltype in ["spherical", "ellipsoidal"]:
            norm_matrix = np.zeros([3, 3])

        self.stress_normalization_matrix = norm_matrix

        # Set info for on-diagonal stress symmetrization
        if self.ltype in ["tetragonal", "trigonal", "hexagonal"]:
            self.stress_indices = [(0, 0), (1, 1)]

        elif self.ltype in ["cubic"]:
            self.stress_indices = [(0, 0), (1, 1), (2, 2)]

        else:
            self.stress_indices = []

        # Set values for the matrix
        if matrix is None:
            self.reset_matrix()
        else:
            self.set_matrix(matrix)

        # Set tolerance
        if self.ltype in ["triclinic"]:
            self.a_tol = 15.0
        else:
            self.a_tol = 9.9
        self._get_dof()

    def _get_dof(self):
        """
        get the number of degree of freedom
        """
        if self.ltype in ["triclinic"]:
            self.dof = 6
        elif self.ltype in ["monoclinic"]:
            self.dof = 4
        elif self.ltype in ["orthorhombic"]:
            self.dof = 3
        elif self.ltype in ["tetragonal", "hexagonal", "trigonal"]:
            self.dof = 2
        else:
            self.dof = 1

    def get_bounds(self, min_vec=2.0, max_vec=50.0, min_ang=30, max_ang=150):
        """
        get the number of degree of freedom
        """
        if self.ltype in ["triclinic"]:
            self.bounds = [(min_vec, max_vec), (min_vec, max_vec), (min_vec, max_vec),
                           (min_ang, max_ang), (min_ang, max_ang), (min_ang, max_ang)]
        elif self.ltype in ["monoclinic"]:
            self.bounds = [(min_vec, max_vec), (min_vec, max_vec), (min_vec, max_vec),
                           (min_ang, max_ang)]
        elif self.ltype in ["orthorhombic"]:
            self.bounds = [(min_vec, max_vec), (min_vec, max_vec), (min_vec, max_vec)]
        elif self.ltype in ["tetragonal", "hexagonal", "trigonal"]:
            self.bounds = [(min_vec, max_vec), (min_vec, max_vec)]
        else:
            self.bounds = [(min_vec, max_vec)]
        return self.bounds

    @classmethod
    def get_dofs(self, ltype):
        """
        get the number of degree of freedom
        """
        # if ltype is None: ltype = self.ltype

        if ltype in ["triclinic"]:
            dofs = [3, 3]
        elif ltype in ["monoclinic"]:
            dofs = [3, 1]
        elif ltype in ["orthorhombic"]:
            dofs = [3, 0]
        elif ltype in ["tetragonal", "hexagonal", "trigonal"]:
            dofs = [2, 0]
        else:
            dofs = [1, 0]
        return dofs

    def copy(self):
        """
        simply copy the structure
        """
        from copy import deepcopy

        return deepcopy(self)

    def get_lengths(self):
        mat = create_matrix(self.PBC, True)
        mat = np.dot(mat, self.matrix)
        return mat, np.linalg.norm(mat, axis=1)

    def scale(self, factor=1.1):

        matrix = self.matrix
        return Lattice.from_matrix(matrix * factor, ltype=self.ltype)

    def get_permutation_matrices(self):
        """
        Return the possible permutation matrices that donot violate the symmetry
        """
        if self.ltype in ["monoclinic"]:  # permutation between a and c
            return np.array(
                [
                    [[1, 0, 0], [0, 1, 0], [0, 0, 1]],  # self
                    [[0, 0, 1], [0, 1, 0], [1, 0, 0]],  # a-c
                ]
            )
        elif self.ltype in ["triclinic"]:
            return np.array(
                [
                    [[1, 0, 0], [0, 1, 0], [0, 0, 1]],  # self
                    [[1, 0, 0], [0, 0, 1], [0, 1, 0]],  # b-c
                    [[0, 0, 1], [0, 1, 0], [1, 0, 0]],  # a-c
                    [[0, 1, 0], [1, 0, 0], [0, 0, 1]],  # a-b
                ]
            )
        else:
            return [np.eye(3)]

    def get_transformation_matrices(self):
        """
        Return possible transformation matrices that donot violate the symmetry
        """
        if self.ltype in ["monoclinic"]:
            return np.array(
                [
                    [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [1, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [-1, 0, 1]],
                    [[1, 0, 1], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, -1], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [0, -1, 0], [0, 0, -1]],  # change angle
                    # [[-1,0,0],[0,1,0],[0,0,1]], #change angle
                ]
            )

        elif self.ltype in ["triclinic"]:
            return np.array(
                [
                    [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [1, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [-1, 0, 1]],
                    [[1, 0, 1], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, -1], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [0, 1, 1]],
                    [[1, 0, 0], [0, 1, 1], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [0, -1, 1]],
                    [[1, 0, 0], [0, 1, -1], [0, 0, 1]],
                    [[1, 1, 0], [0, 1, 0], [0, 0, 1]],
                    [[1, -1, 0], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [1, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [-1, 1, 0], [0, 0, 1]],
                    # [[-1,0,0],[0,-1,0],[0,0,1]],
                    # [[1,0,0],[0,-1,0],[0,0,-1]],
                    # [[-1,0,0],[0,1,0],[0,0,-1]],
                    [[-1, 0, 0], [0, 1, 0], [0, 0, 1]],
                    [[1, 0, 0], [0, -1, 0], [0, 0, 1]],
                    [[1, 0, 0], [0, 1, 0], [0, 0, -1]],
                ]
            )
        else:
            return [np.eye(3)]

    def search_transformations(self, lat_ref, d_tol=1.0, f_tol=0.1):
        """
        search the closest match to the reference lattice object

        Args:
            lat_ref: reference lattice object
            d_tol: tolerance in angle
            f_tol:
            a_tol:

        Returns:
            a two steps of transformation matrix if the match is possible
        """
        # Find all possible permutation and transformation matrices
        trans1 = self.get_permutation_matrices()
        trans2 = self.get_transformation_matrices()
        tols = np.zeros([len(trans2) * len(trans1), 3])
        trans = []
        switchs = []

        count = 0
        for _i, tran1 in enumerate(trans1):
            lat0 = self.transform(tran1)
            for _j, tran2 in enumerate(trans2):
                tmp = np.dot(tran2, lat0.matrix)
                try:
                    # print("Start", np.linalg.det(tmp))
                    lat2 = Lattice.from_matrix(tmp, ltype=self.ltype)
                    # print("End", np.linalg.det(lat2.matrix))
                    d_tol1, f_tol1, a_tol1, switch = lat2.get_diff(lat_ref)
                    # print(d_tol1, f_tol1, a_tol1, switch)
                except:
                    d_tol1, f_tol1, a_tol1, switch = 10, 1.0, 90, None
                tols[count] = [d_tol1, f_tol1, a_tol1]
                trans.append([tran1, tran2])
                switchs.append(switch)
                count += 1

        trans_good = []
        tols_good = []
        for id in range(len(tols)):
            if (tols[id, 0] < d_tol or tols[id, 1] < f_tol) and tols[id, 2] < self.a_tol:
                if switchs[id]:
                    trans[id].extend([[[1, 0, 0], [0, -1, 0], [0, 0, -1]]])
                # print(tols[id], len(trans[id]))
                trans_good.append(trans[id])
                tols_good.append(tols[id])

        return trans_good, tols_good

    def search_transformation(self, lat_ref, d_tol=1.0, f_tol=0.1):
        """
        search the closest match to the reference lattice object

        Args:
            lat_ref: reference lattice object
            d_tol: tolerance in angle
            f_tol:
            a_tol:

        Returns:
            a two steps of transformation matrix if the match is possible
        """
        # Find all possible permutation and transformation matrices
        trans1 = self.get_permutation_matrices()
        trans2 = self.get_transformation_matrices()

        tols = np.zeros([len(trans2) * len(trans1) + 1, 3])
        trans = []
        switchs = []

        # Check it self
        d_tol1, f_tol1, a_tol1, switch = self.get_diff(lat_ref)
        tols[0] = [d_tol1, f_tol1, a_tol1]
        switchs.append(switch)
        trans.append([np.eye(3)])

        count = 0
        for _i, tran1 in enumerate(trans1):
            lat0 = self.transform(tran1)
            for _j, tran2 in enumerate(trans2):
                count += 1
                tmp = np.dot(tran2, lat0.matrix)
                try:
                    # print(i, j, self.ltype)
                    lat2 = Lattice.from_matrix(tmp, ltype=self.ltype)
                    d_tol1, f_tol1, a_tol1, switch = lat2.get_diff(lat_ref)
                    # print(d_tol1, f_tol1, a_tol1, switch)
                except:
                    d_tol1, f_tol1, a_tol1, switch = 10, 1.0, 90, None
                tols[count] = [d_tol1, f_tol1, a_tol1]
                trans.append([tran1, tran2])
                switchs.append(switch)

        # QZ: needs to figure out a better way to select the best
        rms = tols.sum(axis=1)
        ids = np.argsort(rms)
        id = ids[0]
        # print(tols, rms)
        # print(id, switchs[id])
        if abs(rms[ids[0]] - rms[ids[1]]) < 1e-3 and switchs[ids[0]] and not switchs[ids[1]]:
            id = ids[1]
            # print("change id 1", id)
        if id != 0 and abs(rms[0] - rms[id]) < 1.0:
            # print("change id 2", id, rms[0], rms[id])
            id = 0

        if (tols[id, 0] < d_tol or tols[id, 1] < f_tol) and tols[id, 2] < self.a_tol:
            if switchs[id]:
                trans[id].append([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
                return trans[id], tols[id]
            else:
                return trans[id], tols[id]
        else:
            # print("===================================Cannot match:", tols[id])
            return None, None

    def optimize_once(self, reset=False):
        """
        Optimize the lattice's inclination angles
        """
        opt = False
        trans = self.get_transformation_matrices()
        if len(trans) > 1:
            diffs = []
            for tran in trans:
                cell_new = np.dot(tran, self.matrix)
                try:
                    lat_new = Lattice.from_matrix(cell_new, ltype=self.ltype)
                    diffs.append(lat_new.get_worst_angle())
                except:
                    diffs.append(100)
            id = np.array(diffs).argmin()
            if id > 0 and diffs[id] < diffs[0] - 0.01:
                opt = True
                tran = trans[id]
                cell = np.dot(tran, self.matrix)
                lat = Lattice.from_matrix(cell, ltype=self.ltype, reset=reset)
                return lat, tran, opt
        return self, np.eye(3), opt

    def get_worst_angle(self):
        """
        return the worst inclination angle difference w.r.t 90 degree
        """
        return np.max(abs(np.array([self.alpha, self.beta, self.gamma]) - np.pi / 2))

    def optimize_multi(self, iterations=5):
        """
        Optimize the lattice if the cell has a bad inclination angles

        Args:
            iterations: maximum number of iterations
            force: whether or not do the early termination

        Returns:
            the optimized lattice
        """
        lattice = self
        trans_matrices = []
        for _i in range(iterations):
            lattice, trans, opt = lattice.optimize_once(reset=True)
            if opt:
                trans_matrices.append(trans)
            else:
                break
        return lattice, trans_matrices

    def standardize(self):
        """
        Force the angle to be smaller than 90 degree
        """
        change = False
        if self.ltype in ["monoclinic"]:
            if self.beta > np.pi / 2:
                self.beta = np.pi - self.beta
                change = True
        elif self.ltype in ["triclinic"]:
            if self.alpha > np.pi / 2:
                self.alpha = np.pi - self.alpha
                change = True
            if self.beta > np.pi / 2:
                self.beta = np.pi - self.beta
                change = True
            if self.gamma > np.pi / 2:
                self.gamma = np.pi - self.gamma
                change = True

        if change:
            para = (self.a, self.b, self.c, self.alpha, self.beta, self.gamma)
            self.matrix = para2matrix(para)

    def transform(self, trans_mat=None, reset=False):
        """
        Optimize the lattice's inclination angles
        If reset is False, may return negative lattice
        """
        if trans_mat is None:
            trans_mat = np.eye(3)

        if not isinstance(trans_mat, np.ndarray):
            trans_mat = np.array(trans_mat)

        cell = np.dot(trans_mat, self.matrix)
        return Lattice.from_matrix(cell, ltype=self.ltype, reset=reset)

    def transform_multi(self, trans, reset=True):
        """
        Optimize the lattice's inclination angles
        """
        lat = self
        for tran in trans:
            lat = lat.transform(tran, reset)
        return lat

    def encode(self):
        a, b, c, alpha, beta, gamma = self.get_para(degree=True)
        if self.ltype in ["cubic"]:
            return [a]
        elif self.ltype in ["hexagonal", "trigonal", "tetragonal"]:
            return [a, c]
        elif self.ltype in ["orthorhombic"]:
            return [a, b, c]
        elif self.ltype in ["monoclinic"]:
            return [a, b, c, beta]
        else:
            return [a, b, c, alpha, beta, gamma]

    @classmethod
    def from_1d_representation(self, v, ltype):
        # print('test', v, type(v), len(v), v[0])
        if ltype == "triclinic":
            a, b, c, alpha, beta, gamma = v[0], v[1], v[2], v[3], v[4], v[5]
        elif ltype == "monoclinic":
            a, b, c, alpha, beta, gamma = v[0], v[1], v[2], 90, v[3], 90
        elif ltype == "orthorhombic":
            a, b, c, alpha, beta, gamma = v[0], v[1], v[2], 90, 90, 90
        elif ltype == "tetragonal":
            a, b, c, alpha, beta, gamma = v[0], v[0], v[1], 90, 90, 90
        elif ltype in ["trigonal", "hexagonal"]:
            a, b, c, alpha, beta, gamma = v[0], v[0], v[1], 90, 90, 120
        else:
            a, b, c, alpha, beta, gamma = v[0], v[0], v[0], 90, 90, 90
        try:
            return Lattice.from_para(a, b, c, alpha, beta, gamma, ltype=ltype)
        except:
            print(a, b, c, alpha, beta, gamma, ltype)

    def mutate(self, degree=0.10, frozen=False):
        """
        Mutate the lattice object
        """
        rand = 1 + degree * (self.random_state.random(6) - 0.5)
        a0, b0, c0, alpha0, beta0, gamma0 = self.get_para()
        a = a0 * rand[0]
        b = b0 * rand[1]
        c = c0 * rand[2]
        alpha = np.degrees(alpha0 * rand[3])
        beta = np.degrees(beta0 * rand[4])
        gamma = np.degrees(gamma0 * rand[5])
        ltype = self.ltype

        if self.ltype in ["cubic"]:
            if frozen:
                lat = Lattice.from_para(a0, a0, a0, 90, 90, 90, ltype=ltype)
            else:
                lat = Lattice.from_para(a, a, a, 90, 90, 90, ltype=ltype)
        elif ltype in ["hexagonal", "trigonal"]:
            if frozen:
                lat = Lattice.from_para(a0, a0, c, 90, 90, 120, ltype=ltype)
            else:
                lat = Lattice.from_para(a, a, c, 90, 90, 120, ltype=ltype)
        elif ltype in ["tetragonal"]:
            if frozen:
                lat = Lattice.from_para(a0, a0, c, 90, 90, 90, ltype=ltype)
            else:
                lat = Lattice.from_para(a, a, c, 90, 90, 90, ltype=ltype)
        elif ltype in ["orthorhombic"]:
            lat = Lattice.from_para(a, b, c, 90, 90, 90, ltype=ltype)
        elif ltype in ["monoclinic"]:
            lat = Lattice.from_para(a, b, c, 90, beta, 90, ltype=ltype)
        elif ltype in ["triclinic"]:
            lat = Lattice.from_para(a, b, c, alpha, beta, gamma, ltype=ltype)
        else:
            raise ValueError(f"ltype {ltype:s} is not supported")
        return lat

    def generate_para(self):
        if self.dim == 3:
            return generate_cellpara(self.ltype, self.volume, random_state=self.random_state, **self.kwargs)
        elif self.dim == 2:
            return generate_cellpara_2D(self.ltype, self.volume, random_state=self.random_state, **self.kwargs)
        elif self.dim == 1:
            return generate_cellpara_1D(self.ltype, self.volume, random_state=self.random_state, **self.kwargs)
        elif self.dim == 0:
            return generate_cellpara_0D(self.ltype, self.volume, random_state=self.random_state, **self.kwargs)
        return None

    def generate_matrix(self):
        """
        Generates a 3x3 matrix for a lattice based on the lattice type and volume
        """
        # Try multiple times in case of failure
        for _i in range(10):
            para = self.generate_para()
            if para is not None:
                return para2matrix(para)
        return None

    def get_matrix(self, shape="upper"):
        """
        Returns a 3x3 numpy array representing the lattice vectors.
        """
        return self.matrix

    def get_para(self, degree=False):
        """
        Returns a tuple of lattice parameters.
        """
        if degree:
            return (
                self.a,
                self.b,
                self.c,
                deg * self.alpha,
                deg * self.beta,
                deg * self.gamma,
            )
        else:
            return (self.a, self.b, self.c, self.alpha, self.beta, self.gamma)

    def set_matrix(self, matrix=None):
        if matrix is not None:
            m = np.array(matrix)
            if np.shape(m) == (3, 3):
                self.matrix = m
                self.inv_matrix = np.linalg.inv(m)
            else:
                print(matrix)
                msg = "Error: matrix must be a 3x3 numpy array or list"
                raise ValueError(msg)
        else:
            self.reset_matrix()

        para = matrix2para(self.matrix)
        self.a, self.b, self.c, self.alpha, self.beta, self.gamma = para
        self.volume = np.linalg.det(self.matrix)

    def set_para(self, para=None, radians=False):
        if para is not None:
            if radians is False:
                para[3] *= rad
                para[4] *= rad
                para[5] *= rad
            self.set_matrix(para2matrix(para))
        else:
            self.set_matrix()

    def reset_matrix(self, shape="upper"):
        if self.random:
            success = False
            for _i in range(5):
                m = self.generate_matrix()
                if m is not None:
                    self.matrix = m
                    self.inv_matrix = np.linalg.inv(m)
                    [a, b, c, alpha, beta, gamma] = matrix2para(self.matrix)
                    self.a = a
                    self.b = b
                    self.c = c
                    self.alpha = alpha
                    self.beta = beta
                    self.gamma = gamma
                    success = True
                    break
            if not success:
                msg = "Cannot generate a good matrix"
                raise ValueError(msg)
        else:
            # a small utility to convert the cell shape
            para = matrix2para(self.matrix)
            self.matrix = para2matrix(para, format=shape)
            self.inv_matrix = np.linalg.inv(self.matrix)

    def set_volume(self, volume):
        if self.allow_volume_reset:
            self.volume = volume

    def swap_axis(self, random=False, ids=None):
        """
        For the lattice
        """
        # only applied to triclinic/monoclinic/orthorhombic
        if self.ltype in ["triclinic", "orthorhombic", "Orthorhombic"]:
            allowed_ids = [
                [0, 1, 2],
                [1, 0, 2],
                [0, 2, 1],
                [2, 1, 0],
                [1, 2, 0],
                [2, 0, 1],
            ]

        elif self.ltype in ["monoclinic"]:
            if abs(self.beta - 90 * rad) > 1e-3:
                allowed_ids = [[0, 1, 2], [2, 1, 0]]
            else:
                allowed_ids = [
                    [0, 1, 2],
                    [1, 0, 2],
                    [0, 2, 1],
                    [2, 1, 0],
                    [1, 2, 0],
                    [2, 0, 1],
                ]
        else:
            allowed_ids = [[0, 1, 2]]

        if random:
            ids = self.random_state.choice(allowed_ids)
        else:
            if ids not in allowed_ids:
                print(ids)
                raise ValueError("the above swap is not allowed in " + self.ltype)

        (a, b, c, alpha, beta, gamma) = self.get_para()
        alpha, beta, gamma = alpha * deg, beta * deg, gamma * deg
        if ids is None:
            return self
        elif ids == [1, 0, 2]:  # a->b
            return self.from_para(b, a, c, beta, alpha, gamma, self.ltype)
        elif ids == [2, 1, 0]:  # a->c
            return self.from_para(c, b, a, gamma, beta, alpha, self.ltype)
        elif ids == [0, 2, 1]:  # b-c
            return self.from_para(a, c, b, alpha, gamma, beta, self.ltype)
        elif ids == [2, 0, 1]:
            return self.from_para(c, a, b, gamma, alpha, beta, self.ltype)
        elif ids == [1, 2, 0]:
            return self.from_para(b, c, a, beta, gamma, alpha, self.ltype)
        else:
            return self

    def swap_angle(self, random=True, ids=None):
        # only applied to triclinic/monoclinic #/hexagonal
        """
        If the angle is not 90. There will be two equivalent versions
        e.g., 80 and 100.
        """
        if self.ltype in ["monoclinic"]:
            allowed_ids = ["beta", "No"]
        elif self.ltype in ["triclinic"]:
            allowed_ids = ["alpha", "beta", "gamma", "No"]
        else:
            allowed_ids = ["No"]

        if random:
            ids = self.random_state.choice(allowed_ids)
        else:
            if ids not in allowed_ids:
                print(ids)
                raise ValueError("the above swap is not allowed in " + self.ltype)

        (a, b, c, alpha, beta, gamma) = self.get_para()
        alpha, beta, gamma = alpha * deg, beta * deg, gamma * deg
        if ids is None:
            return self
        elif ids == "alpha":
            return self.from_para(a, b, c, 180 - alpha, beta, gamma, self.ltype)
        elif ids == "beta":
            return self.from_para(a, b, c, alpha, 180 - beta, gamma, self.ltype)
        elif ids == "gamma":
            return self.from_para(a, b, c, alpha, beta, 180 - gamma, self.ltype)
        else:
            return self

    def add_vacuum(self, coor, frac=True, vacuum=15, PBC=None):
        """
        Adds space above and below a 2D or 1D crystal.

        Args:
            coor: the relative coordinates of the crystal
            vacuum: the amount of space, in Angstroms, to add above and below
            PBC: A periodic boundary condition list,
                Ex: [1,1,1] -> full 3d periodicity, [0,0,1] -> periodicity
                along the z axis

        Returns:
            The transformed lattice and coordinates after the vacuum is added
        """
        if PBC is None:
            PBC = [0, 0, 0]
        matrix = self.matrix
        absolute_coords = np.dot(coor, matrix) if frac else coor

        for i, a in enumerate(PBC):
            if not a:
                ratio = 1 + vacuum / np.linalg.norm(matrix[i])
                matrix[i] *= ratio
                absolute_coords[:, i] += vacuum / 2
        coor = np.dot(absolute_coords, np.linalg.inv(matrix)) if frac else absolute_coords
        return matrix, coor

    def generate_point(self):
        if self.ltype in ["spherical", "ellipsoidal"]:
            # Choose a point within an octant of the unit sphere using Marsaglia's method
            # Generate 3D vector from normal distribution
            vec = self.random_state.normal(0, 1, 3)
            # Generate random radius
            r = self.random_state.random() ** (1 / 3)
            # Normalize and scale
            point = r * vec / np.linalg.norm(vec)
        else:
            point = self.random_state.random(3)

            for i, a in enumerate(self.PBC):
                if not a:
                    if self.ltype in ["hexagonal", "trigonal"]:
                        point[i] *= 1.0 / np.sqrt(3.0)
                    else:
                        point[i] -= 0.5
        return point

    @classmethod
    def from_para(
        self,
        a,
        b,
        c,
        alpha,
        beta,
        gamma,
        ltype="triclinic",
        radians=False,
        PBC=None,
        factor=1.0,
        force_symmetry=False,
        **kwargs,
    ):
        """
        Creates a Lattice object from 6 lattice parameters. Additional keyword
        arguments  are available. Unless specified by the keyword random=True,
        does not create a new matrix upon calling reset_matrix. This allows
        for generation of random crystals with a specific choice of unit cell.

        Args:
            a: The length (in Angstroms) of the unit cell vectors
            b: The length (in Angstroms) of the unit cell vectors
            c: The length (in Angstroms) of the unit cell vectors
            alpha: the angle (in degrees) between the b and c vectors
            beta: the angle (in degrees) between the a and c vectors
            gamma: the angle (in degrees) between the a and b vectors
            ltype: the lattice type ("cubic, tetragonal, etc."). Also available
                are "spherical", which confines generated points to lie within a
                sphere, and "ellipsoidal", which confines generated points to lie
                within an ellipse (oriented about the z axis)
            radians: whether or not to use radians (instead of degrees) for the
                lattice angles
            PBC: A periodic boundary condition list, where 1 means periodic,
                0 means not periodic.
                Ex: [1,1,1] -> full 3d periodicity, [0,0,1] -> periodicity along
                the z axis
            kwargs: various values which may be defined. If none are defined,
                random ones will be generated. Values will be passed to
                generate_lattice. Options include:
                area: The cross-sectional area (in Angstroms squared). Only used
                    to generate 1D crystals
                thickness: The unit cell's non-periodic thickness (in Angstroms).
                    Only used to generate 2D crystals
                unique_axis: The unique axis for certain symmetry (and especially
                    layer) groups. Because the symmetry operations are not also
                    transformed, you should use the default values for random
                    crystal generation
                random: If False, keeps the stored values for the lattice geometry
                    even upon applying reset_matrix. To alter the matrix,
                    use set_matrix() or set_para
                'unique_axis': the axis ('a', 'b', or 'c') which is unique
                'min_l': the smallest allowed cell vector.
                'mid_l': the second smallest allowed cell vector.
                'max_l': the third smallest allowed cell vector.

        Returns:
            a Lattice object with the specified parameters
        """
        if PBC is None:
            PBC = [1, 1, 1]
        try:
            cell_matrix = para2matrix((a, b, c, alpha, beta, gamma), radians=radians)
            cell_matrix *= factor
        except Exception as err:
            msg = "Error: invalid cell parameters for lattice."
            raise ValueError(msg) from err

        if force_symmetry:
            return Lattice.from_matrix(cell_matrix, ltype=ltype)
        else:
            volume = np.linalg.det(cell_matrix)
            # Initialize a Lattice instance
            l = Lattice(ltype, volume, PBC=PBC, **kwargs)
            l.a, l.b, l.c = factor * a, factor * b, factor * c
            l.alpha, l.beta, l.gamma = alpha * rad, beta * rad, gamma * rad
            l.matrix = cell_matrix
            l.inv_matrix = np.linalg.inv(cell_matrix)
            l.ltype = ltype
            l.volume = volume
            l.random = False
            l.allow_volume_reset = False
            return l

    @classmethod
    def from_matrix(
        self,
        matrix,
        reset=True,
        shape="upper",
        ltype="triclinic",
        PBC=None,
        **kwargs,
    ):
        """
        Creates a Lattice object from a 3x3 cell matrix. Additional keywords
        are available. Unless specified by the keyword random=True, does not
        create a new matrix upon calling reset_matrix. This allows for a random
        crystals with a specific choice of unit cell.

        Args:
            matrix: a 3x3 real matrix (numpy array or nested list) for the cell
            ltype: the lattice type ("cubic, tetragonal, etc."). Also can be
                - "spherical", confines points to lie within a sphere,
                - "ellipsoidal", points to lie within an ellipsoid (about z axis)
            PBC: A periodic boundary condition list, where 1 is periodic
                Ex: [1,1,1] -> full 3d periodicity, [0,0,1] -> periodicity at z.
            kwargs: various values which may be defined. Random ones if None
                Values will be passed to generate_lattice. Options include:
                `area: The cross-sectional area (in Ang^2) for 1D crystals
                `thickness`: The cell's thickness (in Ang) for 2D crystals
                `unique_axis`: The unique axis for layer groups.
                `random`: If False, keeps the stored values for the lattice
                geometry even applying reset_matrix. To alter the matrix,
                use `set_matrix()` or `set_para`
                'unique_axis': the axis ('a', 'b', or 'c') which is unique.
                'min_l': the smallest allowed cell vector.
                'mid_l': the second smallest allowed cell vector.
                'max_l': the third smallest allowed cell vector.

        Returns:
            a Lattice object with the specified parameters
        """
        if PBC is None:
            PBC = [1, 1, 1]

        m = np.array(matrix)
        if np.shape(m) != (3, 3):
            print(matrix)
            msg = "Error: matrix must be a 3x3 numpy array or list"
            raise ValueError(msg)

        [a, b, c, alpha, beta, gamma] = matrix2para(m)

        # symmetrize the lattice
        if reset:
            if ltype in ["cubic", "Cubic"]:
                a = b = c = (a + b + c) / 3
                alpha = beta = gamma = np.pi / 2
            elif ltype in ["hexagonal", "trigonal", "Hexagonal", "Trigonal"]:
                a = b = (a + b) / 2
                alpha = beta = np.pi / 2
                gamma = np.pi * 2 / 3
            elif ltype in ["tetragonal", "Tetragonal"]:
                a = b = (a + b) / 2
                alpha = beta = gamma = np.pi / 2
            elif ltype in ["orthorhombic", "Orthorhombic"]:
                alpha = beta = gamma = np.pi / 2
            elif ltype in ["monoclinic", "Monoclinic"]:
                alpha = gamma = np.pi / 2

            # reset matrix according to the symmetry
            m = para2matrix([a, b, c, alpha, beta, gamma], format=shape)

        # Initialize a Lattice instance
        volume = np.linalg.det(m)
        l = Lattice(ltype, volume, m, PBC=PBC, **kwargs)
        l.a, l.b, l.c = a, b, c
        l.alpha, l.beta, l.gamma = alpha, beta, gamma
        l.matrix = m
        l.inv_matrix = np.linalg.inv(m)
        l.ltype = ltype
        l.volume = volume
        l.random = False
        l.allow_volume_reset = False
        return l

    def is_valid_matrix(self):
        """
        check if the cell parameter is reasonable or not
        """

        try:
            paras = [self.a, self.b, self.c, self.alpha, self.beta, self.gamma]
            para2matrix(paras)
            return True
        except:
            return False

    def is_valid_lattice(self, tol=1e-3):
        ltype = self.ltype.lower()

        def check_angles(angles):
            return all(abs(angle - np.pi / 2) <= tol for angle in angles)

        if ltype == "cubic":
            return (
                abs(self.a - self.b) <= tol
                and abs(self.a - self.c) <= tol
                and check_angles([self.alpha, self.beta, self.gamma])
            )

        elif ltype in ["hexagonal", "trigonal"]:
            return (
                abs(self.a - self.b) <= tol
                and check_angles([self.alpha, self.beta])
                and abs(self.gamma - 2 / 3 * np.pi) <= tol
            )

        elif ltype == "tetragonal":
            return abs(self.a - self.b) <= tol and check_angles([self.alpha, self.beta, self.gamma])

        elif ltype == "orthorhombic":
            return check_angles([self.alpha, self.beta, self.gamma])

        elif ltype == "monoclinic":
            return check_angles([self.alpha, self.gamma])

        return True

    def check_mismatch(self, trans, l_type, tol=1.0, a_tol=10):
        """
        check if the lattice mismatch is big after a transformation
        This is mostly used in supergroup function
        QZ: to fix ===============

        Args:
            trans: 3*3 matrix
            l_type: lattice_type like orthrhombic
            tol: tolerance in a, b, c
            a_tol: tolerance in alpha, beta, gamma

        Returns:
            True or False
        """
        matrix = np.dot(trans.T, self.matrix)
        l1 = Lattice.from_matrix(matrix)
        l2 = Lattice.from_matrix(matrix, ltype=l_type)
        (a1, b1, c1, alpha1, beta1, gamma1) = l1.get_para(degree=True)
        (a2, b2, c2, alpha2, beta2, gamma2) = l2.get_para(degree=True)
        abc_diff = np.abs(np.array([a2 - a1, b2 - b1, c2 - c1])).max()
        ang_diff = np.abs(np.array([alpha2 - alpha1, beta2 - beta1, gamma2 - gamma1])).max()
        return not (abc_diff > tol or ang_diff > a_tol)

    def get_diff(self, l_ref):
        """
        get the difference in length, angle, and check if switch is needed
        """
        (a1, b1, c1, alpha1, beta1, gamma1) = self.get_para(degree=True)
        (a2, b2, c2, alpha2, beta2, gamma2) = l_ref.get_para(degree=True)
        abc_diff = np.abs(np.array([a2 - a1, b2 - b1, c2 - c1])).max()
        abc_f_diff = np.abs(np.array([(a2 - a1) / a1, (b2 - b1) / b1, (c2 - c1) / c1])).max()
        ang_diff1 = abs(alpha1 - alpha2) + abs(beta1 - beta2) + abs(gamma1 - gamma2)
        ang_diff2 = abs(alpha1 - alpha2)
        ang_diff2 += abs(abs(beta1 - 90) - abs(beta2 - 90))
        ang_diff2 += abs(gamma1 - gamma2)
        # print(abc_diff, abc_f_diff, ang_diff1, ang_diff2, self.ltype)
        if ang_diff1 < ang_diff2 + 0.01:
            return abc_diff, abc_f_diff, ang_diff1, False
        else:
            if self.ltype == "monoclinic":
                return abc_diff, abc_f_diff, ang_diff2, True
            else:
                return abc_diff, abc_f_diff, ang_diff2, False

    def __str__(self):
        return (
            f"{self.a:8.4f}, {self.b:8.4f}, {self.c:8.4f}, {self.alpha * deg:8.4f}, "
            f"{self.beta * deg:8.4f}, {self.gamma * deg:8.4f}, {self.ltype!s:s}"
        )

    def __repr__(self):
        return str(self)

    def find_transition_to_orthoslab(self, c=(0, 0, 1), a=(1, 0, 0), m=5):
        """
        Create the slab model with an approximate orthogonal box shape
        """
        from pyxtal.plane import has_reduction

        tol = 1e-3
        direction = np.array(c)

        # find the simplest a-direction
        if np.dot(np.array(a), direction) < tol:
            a_hkl = np.array(a)
        else:
            a_hkls = []
            for h in range(-m, m + 1):
                for k in range(-m, m + 1):
                    for l in range(-m, m + 1):
                        hkl = np.array([h, k, l])
                        if (
                            ([h, k, l] != [0, 0, 0])
                            and (not has_reduction(hkl))
                            and (abs(np.dot(hkl, direction)) < tol)
                        ):
                            a_hkls.append(hkl)
            a_hkls = np.array(a_hkls)  # ; print(a_hkls)
            a_hkl = a_hkls[np.argmin(np.abs(a_hkls).sum(axis=1))]
        a_vector = np.dot(a_hkl, self.matrix)
        # print('a_hkl', a_hkl)

        # find the simplest b-direction
        b_hkl = None
        min_angle_ab = float("inf")
        for h in range(-m, m + 1):
            for k in range(-m, m + 1):
                for l in range(-m, m + 1):
                    hkl = np.array([h, k, l])
                    if ([h, k, l] != [0, 0, 0]) and (not has_reduction(hkl)) and (abs(np.dot(hkl, direction)) < tol):
                        vector = np.dot(hkl, self.matrix)
                        angle1 = angle(vector, a_vector, radians=False)
                        if abs(90 - angle1) < min_angle_ab:
                            min_angle_ab = abs(90 - angle1)
                            b_hkl = hkl
                            b_vector = vector

        # print('b_hkl', b_hkl, min_angle_ab)
        # change the sign
        if abs(angle(np.cross(a_hkl, b_hkl), direction)) > tol:
            b_hkl *= -1
            b_vector *= -1

        ## update the c_direction
        ab_plane = np.cross(a_vector, b_vector)  # ; print('ab_plane', ab_plane)
        c_hkl = None
        min_angle_c = float("inf")
        for h in range(-m, m + 1):
            for k in range(-m, m + 1):
                for l in range(-m, m + 1):
                    hkl = np.array([h, k, l])
                    if [h, k, l] != [0, 0, 0] and not has_reduction(hkl):
                        vector = np.dot(hkl, self.matrix)
                        angle1 = angle(vector, ab_plane, radians=False)
                        # print(hkl, angle)
                        if abs(angle1) < abs(min_angle_c):
                            min_angle_c = angle1
                            c_hkl = hkl

        # print(a_hkl, b_hkl, c_hkl)
        return np.vstack([a_hkl, b_hkl, c_hkl])

    def apply_transformation(self, trans):
        """
        Optimize the lattice's inclination angles
        """
        cell_new = np.dot(trans, self.matrix)
        return Lattice.from_matrix(cell_new)


def generate_cellpara(
    ltype,
    volume,
    minvec=1.2,
    minangle=np.pi / 6,
    max_ratio=10.0,
    min_special=None,
    maxattempts=100,
    random_state: None | int | Generator = None,
    **kwargs,
):
    """
    Generates the cell parameter (a, b, c, alpha, beta, gamma) according
    to the space group symmetry and number of atoms. If the spacegroup
    has centering, we will transform to conventional cell setting. If the
    generated lattice does not meet the minimum angle and vector
    requirements, we try to generate a new one, up to maxattempts times.

    Args:
        volume: volume of the conventional unit cell
        minvec: minimum allowed lattice vector length (among a, b, and c)
        minangle: minimum allowed lattice angle (among alpha, beta, and gamma)
        max_ratio: largest allowed ratio of two lattice vector lengths
        maxattempts: the maximum number of attempts for generating a lattice
        kwargs: a dictionary of optional values. These include:
            'unique_axis': the axis ('a', 'b', or 'c') which is unique.
            'min_l': the smallest allowed cell vector.
            'mid_l': the second smallest allowed cell vector.
            'max_l': the third smallest allowed cell vector.

    Returns:
        a 6-length array representing the lattice of the unit cell. If
        generation fails, outputs a warning message and returns empty
    """
    if isinstance(random_state, Generator):
        random_state = random_state.spawn(1)[0]
    else:
        random_state = np.random.default_rng(random_state)

    min_special = kwargs.get("min_special", min_special)  # ; print("min_special", min_special)
    maxangle = np.pi - minangle
    for _n in range(maxattempts):
        # Triclinic
        # if sg <= 2:
        if ltype == "triclinic":
            # Derive lattice constants from a random matrix
            mat = random_shear_matrix(width=0.2, random_state=random_state)
            a, b, c, alpha, beta, gamma = matrix2para(mat)
            x = np.sqrt(
                1
                - np.cos(alpha) ** 2
                - np.cos(beta) ** 2
                - np.cos(gamma) ** 2
                + 2 * (np.cos(alpha) * np.cos(beta) * np.cos(gamma))
            )
            vec = random_vector(random_state=random_state)
            abc = volume / x
            xyz = vec[0] * vec[1] * vec[2]
        # Monoclinic
        elif ltype in ["monoclinic"]:
            alpha, gamma = np.pi / 2, np.pi / 2
            beta = gaussian_random_variable(minangle, maxangle, random_state=random_state)
            x = np.sin(beta)
            vec = random_vector(random_state=random_state)
            xyz = vec[0] * vec[1] * vec[2]
            abc = volume / x
            xyz = vec[0] * vec[1] * vec[2]
        # Orthorhombic
        # elif sg <= 74:
        elif ltype in ["orthorhombic"]:
            alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 2
            x = 1
            vec = random_vector(random_state=random_state)
            xyz = vec[0] * vec[1] * vec[2]
            abc = volume / x
        # Tetragonal
        # elif sg <= 142:
        elif ltype in ["tetragonal"]:
            alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 2
            x = 1
            vec = random_vector(random_state=random_state)
            c = vec[2] / (vec[0] * vec[1]) * np.cbrt(volume / x)
            a = b = np.sqrt((volume / x) / c)
        # Trigonal/Rhombohedral/Hexagonal
        # elif sg <= 194:
        elif ltype in ["hexagonal", "trigonal"]:
            alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 3 * 2
            x = np.sqrt(3.0) / 2.0
            vec = random_vector(random_state=random_state)
            c = vec[2] / (vec[0] * vec[1]) * np.cbrt(volume / x)
            a = b = np.sqrt((volume / x) / c)
        # Cubic
        # else:
        elif ltype in ["cubic"]:
            alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 2
            s = (volume) ** (1.0 / 3.0)
            a, b, c = s, s, s

        # resort a/b/c if min_special is not None for mol. xtals
        if ltype in ["triclinic", "monoclinic", "orthorhombic"]:
            vec *= np.cbrt(abc) / np.cbrt(xyz)
            if min_special is not None:
                ax = random_state.choice([0, 1, 2])
                if vec[ax] < min_special:
                    coef = random_state.uniform(0.8, 1.2) * min_special / vec[ax]
                    for i in range(3):
                        if i == ax:
                            vec[i] *= coef
                        else:
                            vec[i] /= np.sqrt(coef)
            [a, b, c] = vec

        # Check that lattice meets requirements
        maxvec = (a * b * c) / (minvec**2)

        # Define limits on cell dimensions
        min_l = kwargs.get("min_l", minvec)
        mid_l = kwargs.get("mid_l", min_l)
        max_l = kwargs.get("max_l", mid_l)
        l_min = min(a, b, c)
        l_max = max(a, b, c)
        for x in (a, b, c):
            if x <= l_max and x >= l_min:
                l_mid = x
        if not (l_min >= min_l and l_mid >= mid_l and l_max >= max_l):
            continue

        if minvec < maxvec:
            # Check minimum Euclidean distances
            smallvec = min(
                a * np.cos(max(beta, gamma)),
                b * np.cos(max(alpha, gamma)),
                c * np.cos(max(alpha, beta)),
            )
            if (
                a > minvec
                and b > minvec
                and c > minvec
                and a < maxvec
                and b < maxvec
                and c < maxvec
                and smallvec < minvec
                and alpha > minangle
                and beta > minangle
                and gamma > minangle
                and alpha < maxangle
                and beta < maxangle
                and gamma < maxangle
                and a / b < max_ratio
                and a / c < max_ratio
                and b / c < max_ratio
                and b / a < max_ratio
                and c / a < max_ratio
                and c / b < max_ratio
            ):
                return np.array([a, b, c, alpha, beta, gamma])

    # If maxattempts tries have been made without success
    msg = f"lattice fails after {maxattempts:d} cycles"
    msg += f"for volume {volume:.2f}"
    raise VolumeError(msg)
    # return


def generate_cellpara_2D(
    ltype,
    volume,
    thickness=None,
    minvec=1.2,
    minangle=np.pi / 6,
    max_ratio=10.0,
    maxattempts=100,
    random_state: None | int | Generator = None,
    **kwargs,
):
    """
    Generates the cell parameter (a, b, c, alpha, beta, gamma) according
    to the layer group symmetry and number of atoms. If the layer group
    has centering, we will transform to conventional cell setting. If the
    generated lattice does not meet the minimum angle and vector
    requirements, we try to generate a new one, up to maxattempts times.

    Note: The monoclinic layer groups have different unique axes. Groups 3-7
        have unique axis c, while 8-18 have unique axis a. We use non-periodic
        axis c for all layer groups.

    Args:
        num: International number of the space group
        volume: volume of the lattice
        thickness: 3rd-dimensional thickness of the unit cell. If set to None,
            a thickness is chosen automatically
        minvec: minimum allowed lattice vector length (among a, b, and c)
        minangle: minimum allowed lattice angle (among alpha, beta, and gamma)
        max_ratio: largest allowed ratio of two lattice vector lengths
        maxattempts: the maximum number of attempts for generating a lattice
        kwargs: a dictionary of optional values. These include:
            'unique_axis': the axis ('a', 'b', or 'c') which is unique.
            'min_l': the smallest allowed cell vector.
            'mid_l': the second smallest allowed cell vector.
            'max_l': the third smallest allowed cell vector.

    Returns:
        a 6-length representing the lattice vectors of the unit cell. If
        generation fails, outputs a warning message and returns empty
    """
    if isinstance(random_state, int):
        # NOTE if random_state is an integer make a Generator to ensure randomness
        # downstream that would be lost if integer seed used repeated
        random_state = np.random.default_rng(random_state)

    unique_axis = kwargs.get("unique_axis", "c")
    # Store the non-periodic axis
    NPA = 3
    # Set the unique axis for monoclinic cells
    # if num in range(3, 8): unique_axis = "c"
    # elif num in range(8, 19): unique_axis = "a"
    maxangle = np.pi - minangle
    for _n in range(maxattempts):
        abc = np.ones([3])
        if thickness is None:
            v = random_vector(random_state=random_state)
            thickness1 = np.cbrt(volume) * (v[0] / (v[0] * v[1] * v[2]))
        else:
            thickness1 = max([3.0, thickness])
        abc[NPA - 1] = thickness1
        alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 2
        # Triclinic
        # if num <= 2:
        if ltype == "triclinic":
            mat = random_shear_matrix(width=0.2, random_state=random_state)
            a, b, c, alpha, beta, gamma = matrix2para(mat)
            x = np.sqrt(
                1
                - np.cos(alpha) ** 2
                - np.cos(beta) ** 2
                - np.cos(gamma) ** 2
                + 2 * (np.cos(alpha) * np.cos(beta) * np.cos(gamma))
            )
            abc[NPA - 1] = abc[NPA - 1] / x  # scale thickness by outer product of vectors
            ab = volume / (abc[NPA - 1] * x)
            ratio = a / b
            if NPA == 3:
                abc[0] = np.sqrt(ab * ratio)
                abc[1] = np.sqrt(ab / ratio)
            elif NPA == 2:
                abc[0] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)
            elif NPA == 1:
                abc[1] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)

        # Monoclinic
        # elif num <= 18:
        elif ltype == "monoclinic":
            a, b, c = random_vector(random_state=random_state)
            if unique_axis == "a":
                alpha = gaussian_random_variable(minangle, maxangle, random_state=random_state)
                x = np.sin(alpha)
            elif unique_axis == "b":
                beta = gaussian_random_variable(minangle, maxangle, random_state=random_state)
                x = np.sin(beta)
            elif unique_axis == "c":
                gamma = gaussian_random_variable(minangle, maxangle, random_state=random_state)
                x = np.sin(gamma)
            ab = volume / (abc[NPA - 1] * x)
            ratio = a / b
            if NPA == 3:
                abc[0] = np.sqrt(ab * ratio)
                abc[1] = np.sqrt(ab / ratio)
            elif NPA == 2:
                abc[0] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)
            elif NPA == 1:
                abc[1] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)

        # Orthorhombic
        # elif num <= 48:
        elif ltype == "orthorhombic":
            vec = random_vector(random_state=random_state)
            if NPA == 3:
                ratio = abs(vec[0] / vec[1])  # ratio a/b
                abc[1] = np.sqrt(volume / (thickness1 * ratio))
                abc[0] = abc[1] * ratio
            elif NPA == 2:
                ratio = abs(vec[0] / vec[2])  # ratio a/b
                abc[2] = np.sqrt(volume / (thickness1 * ratio))
                abc[0] = abc[2] * ratio
            elif NPA == 1:
                ratio = abs(vec[1] / vec[2])  # ratio a/b
                abc[2] = np.sqrt(volume / (thickness1 * ratio))
                abc[1] = abc[2] * ratio

        # Tetragonal
        # elif num <= 64:
        elif ltype == "tetragonal":
            if NPA == 3:
                abc[0] = abc[1] = np.sqrt(volume / thickness1)
            elif NPA == 2:
                abc[0] = abc[1]
                abc[2] = volume / (abc[NPA - 1] ** 2)
            elif NPA == 1:
                abc[1] = abc[0]
                abc[2] = volume / (abc[NPA - 1] ** 2)

        # Trigonal/Hexagonal
        # elif num <= 80:
        elif ltype in ["hexagonal", "trigonal"]:
            gamma = np.pi / 3 * 2
            x = np.sqrt(3.0) / 2.0
            if NPA == 3:
                abc[0] = abc[1] = np.sqrt((volume / x) / abc[NPA - 1])
            elif NPA == 2:
                abc[0] = abc[1]
                abc[2] = (volume / x)(thickness1**2)
            elif NPA == 1:
                abc[1] = abc[0]
                abc[2] = (volume / x) / (thickness1**2)

        para = np.array([abc[0], abc[1], abc[2], alpha, beta, gamma])

        a, b, c = abc[0], abc[1], abc[2]
        maxvec = (a * b * c) / (minvec**2)

        # Define limits on cell dimensions
        min_l = kwargs.get("min_l", minvec)
        mid_l = kwargs.get("mid_l", min_l)
        max_l = kwargs.get("max_l", mid_l)
        l_min = min(a, b, c)
        l_max = max(a, b, c)
        for x in (a, b, c):
            if x <= l_max and x >= l_min:
                l_mid = x
        if not (l_min >= min_l and l_mid >= mid_l and l_max >= max_l):
            continue

        if minvec < maxvec:
            smallvec = min(
                a * np.cos(max(beta, gamma)),
                b * np.cos(max(alpha, gamma)),
                c * np.cos(max(alpha, beta)),
            )
            if (
                a > minvec
                and b > minvec
                and c > minvec
                and a < maxvec
                and b < maxvec
                and c < maxvec
                and smallvec < minvec
                and alpha > minangle
                and beta > minangle
                and gamma > minangle
                and alpha < maxangle
                and beta < maxangle
                and gamma < maxangle
                and a / b < max_ratio
                and a / c < max_ratio
                and b / c < max_ratio
                and b / a < max_ratio
                and c / a < max_ratio
                and c / b < max_ratio
            ):
                return para

    # If maxattempts tries have been made without success
    msg = f"Cannot get lattice after {maxattempts:d} cycles for volume {volume:.2f}"
    raise VolumeError(msg)


def generate_cellpara_1D(
    ltype,
    volume,
    area=None,
    minvec=1.2,
    minangle=np.pi / 6,
    max_ratio=10.0,
    maxattempts=100,
    random_state: None | int | Generator = None,
    **kwargs,
):
    """
    Generates a cell parameter (a, b, c, alpha, beta, gamma) according to
    the rod group symmetry and number of atoms. If the rod group has centering,
    we will transform to conventional cell setting. If the generated lattice
    does not meet the minimum angle and vector requirements, we try to
    generate a new one, up to maxattempts times.

    Note: The monoclinic Rod groups have different unique axes. Groups 3-7
        have unique axis a, while 8-12 have unique axis c. We use periodic
        axis c for all Rod groups.

    Args:
        num: number of the Rod group
        volume: volume of the lattice
        area: cross-sectional area of the unit cell in Angstroms squared. If
            set to None, a value is chosen automatically
        minvec: minimum allowed lattice vector length (among a, b, and c)
        minangle: minimum allowed lattice angle (among alpha, beta, and gamma)
        max_ratio: largest allowed ratio of two lattice vector lengths
        maxattempts: the maximum number of attempts for generating a lattice
        kwargs: a dictionary of optional values. These include:
            'unique_axis': the axis ('a', 'b', or 'c') which is unique.
            'min_l': the smallest allowed cell vector.
            'mid_l': the second smallest allowed cell vector.
            'max_l': the third smallest allowed cell vector.

    Returns:
        a 6-length array representing the lattice of the unit cell. If
        generation fails, outputs a warning message and returns empty
    """
    if isinstance(random_state, int):
        # NOTE if random_state is an integer make a Generator to ensure randomness
        # downstream that would be lost if integer seed used repeated
        random_state = np.random.default_rng(random_state)

    try:
        unique_axis = kwargs["unique_axis"]
    except:
        unique_axis = "a"
    # Store the periodic axis
    PA = 3
    # Set the unique axis for monoclinic cells
    # if num in range(3, 8): unique_axis = "a"
    # elif num in range(8, 13): unique_axis = "c"
    maxangle = np.pi - minangle
    for _n in range(maxattempts):
        abc = np.ones([3])
        if area is None:
            v = random_vector(random_state=random_state)
            thickness1 = np.cbrt(volume) * (v[0] / (v[0] * v[1] * v[2]))
        else:
            thickness1 = volume / area
        abc[PA - 1] = thickness1
        alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 2
        # Triclinic
        # if num <= 2:
        if ltype == "triclinic":
            mat = random_shear_matrix(width=0.2, random_state=random_state)
            a, b, c, alpha, beta, gamma = matrix2para(mat)
            x = np.sqrt(
                1
                - np.cos(alpha) ** 2
                - np.cos(beta) ** 2
                - np.cos(gamma) ** 2
                + 2 * (np.cos(alpha) * np.cos(beta) * np.cos(gamma))
            )
            abc[PA - 1] = abc[PA - 1] / x  # scale thickness by outer product of vectors
            ab = volume / (abc[PA - 1] * x)
            ratio = a / b
            if PA == 3:
                abc[0] = np.sqrt(ab * ratio)
                abc[1] = np.sqrt(ab / ratio)
            elif PA == 2:
                abc[0] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)
            elif PA == 1:
                abc[1] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)

        # Monoclinic
        # elif num <= 12:
        elif ltype == "monoclinic":
            a, b, c = random_vector(random_state=random_state)
            if unique_axis == "a":
                alpha = gaussian_random_variable(minangle, maxangle, random_state=random_state)
                x = np.sin(alpha)
            elif unique_axis == "b":
                beta = gaussian_random_variable(minangle, maxangle, random_state=random_state)
                x = np.sin(beta)
            elif unique_axis == "c":
                gamma = gaussian_random_variable(minangle, maxangle, random_state=random_state)
                x = np.sin(gamma)
            ab = volume / (abc[PA - 1] * x)
            ratio = a / b
            if PA == 3:
                abc[0] = np.sqrt(ab * ratio)
                abc[1] = np.sqrt(ab / ratio)
            elif PA == 2:
                abc[0] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)
            elif PA == 1:
                abc[1] = np.sqrt(ab * ratio)
                abc[2] = np.sqrt(ab / ratio)

        # Orthorhombic
        # lif num <= 22:
        elif ltype == "orthorhombic":
            vec = random_vector(random_state=random_state)
            if PA == 3:
                ratio = abs(vec[0] / vec[1])  # ratio a/b
                abc[1] = np.sqrt(volume / (thickness1 * ratio))
                abc[0] = abc[1] * ratio
            elif PA == 2:
                ratio = abs(vec[0] / vec[2])  # ratio a/b
                abc[2] = np.sqrt(volume / (thickness1 * ratio))
                abc[0] = abc[2] * ratio
            elif PA == 1:
                ratio = abs(vec[1] / vec[2])  # ratio a/b
                abc[2] = np.sqrt(volume / (thickness1 * ratio))
                abc[1] = abc[2] * ratio

        # Tetragonal
        # elif num <= 41:
        elif ltype == "tetragonal":
            if PA == 3:
                abc[0] = abc[1] = np.sqrt(volume / thickness1)
            elif PA == 2:
                abc[0] = abc[1]
                abc[2] = volume / (abc[PA - 1] ** 2)
            elif PA == 1:
                abc[1] = abc[0]
                abc[2] = volume / (abc[PA - 1] ** 2)

        # Trigonal/Rhombohedral/Hexagonal
        # elif num <= 75:
        elif ltype in ["hexagonal", "trigonal"]:
            gamma = np.pi / 3 * 2
            x = np.sqrt(3.0) / 2.0
            if PA == 3:
                abc[0] = abc[1] = np.sqrt((volume / x) / abc[PA - 1])
            elif PA == 2:
                abc[0] = abc[1]
                abc[2] = (volume / x)(thickness1**2)
            elif PA == 1:
                abc[1] = abc[0]
                abc[2] = (volume / x) / (thickness1**2)

        para = np.array([abc[0], abc[1], abc[2], alpha, beta, gamma])

        a, b, c = abc[0], abc[1], abc[2]
        maxvec = (a * b * c) / (minvec**2)

        # Define limits on cell dimensions
        min_l = kwargs.get("min_l", minvec)
        mid_l = kwargs.get("mid_l", min_l)
        max_l = kwargs.get("max_l", mid_l)
        l_min = min(a, b, c)
        l_max = max(a, b, c)
        for x in (a, b, c):
            if x <= l_max and x >= l_min:
                l_mid = x
        if not (l_min >= min_l and l_mid >= mid_l and l_max >= max_l):
            continue

        if minvec < maxvec:
            smallvec = min(
                a * np.cos(max(beta, gamma)),
                b * np.cos(max(alpha, gamma)),
                c * np.cos(max(alpha, beta)),
            )
            if (
                a > minvec
                and b > minvec
                and c > minvec
                and a < maxvec
                and b < maxvec
                and c < maxvec
                and smallvec < minvec
                and alpha > minangle
                and beta > minangle
                and gamma > minangle
                and alpha < maxangle
                and beta < maxangle
                and gamma < maxangle
                and a / b < max_ratio
                and a / c < max_ratio
                and b / c < max_ratio
                and b / a < max_ratio
                and c / a < max_ratio
                and c / b < max_ratio
            ):
                return para

    # If maxattempts tries have been made without success
    msg = f"Could not get lattice after {maxattempts:d} cycles for volume {volume:.2f}"
    raise VolumeError(msg)


def generate_cellpara_0D(
    ltype,
    volume,
    area=None,
    minvec=1.2,
    max_ratio=10.0,
    maxattempts=100,
    random_state: None | int | Generator = None,
    **kwargs,
):
    """
    Generates a cell parameter (a, b, c, alpha, beta, gamma) according to the
    point group symmetry and number of atoms. If the generated lattice does
    not meet the minimum angle and vector requirements, we try to generate
    a new one, up to maxattempts times.

    Args:
        num: number of the Rod group
        volume: volume of the lattice
        area: cross-sectional area of the unit cell in Angstroms squared. If
            set to None, a value is chosen automatically
        minvec: minimum allowed lattice vector length (among a, b, and c)
        max_ratio: largest allowed ratio of two lattice vector lengths
        maxattempts: the maximum number of attempts for generating a lattice
        kwargs: a dictionary of optional values. Only used for ellipsoidal
            lattices, which pass the value to generate_lattice. They include:
            'unique_axis': the axis ('a', 'b', or 'c') which is unique.
            'min_l': the smallest allowed cell vector.
            'mid_l': the second smallest allowed cell vector.
            'max_l': the third smallest allowed cell vector.

    Returns:
        a 3x3 matrix representing the lattice vectors of the unit cell. If
        generation fails, outputs a warning message and returns empty
    """
    if isinstance(random_state, int):
        # NOTE if random_state is an integer make a Generator to ensure randomness
        # downstream that would be lost if integer seed used repeated
        random_state = np.random.default_rng(random_state)

    if ltype == "spherical":
        # Use a cubic lattice with altered volume
        a = b = c = np.cbrt((3 * volume) / (4 * np.pi))
        alpha = beta = gamma = 0.5 * np.pi
        return np.array([a, b, c, alpha, beta, gamma])
    if ltype == "ellipsoidal":
        # Use a matrix with only on-diagonal elements, with a = b
        alpha, beta, gamma = np.pi / 2, np.pi / 2, np.pi / 2
        x = (4.0 / 3.0) * np.pi
        for _numattempts in range(maxattempts):
            vec = random_vector(random_state=random_state)
            c = vec[2] / (vec[0] * vec[1]) * np.cbrt(volume / x)
            a = b = np.sqrt((volume / x) / c)
            if (a / c < 10.0) and (c / a < 10.0):
                return np.array([a, b, c, alpha, beta, gamma])

    # If maxattempts tries have been made without success
    msg = f"Cannot get lattice after {maxattempts:d} cycles for volume {volume:.2f}"
    raise VolumeError(msg)


def matrix2para(matrix, radians=True):
    """
    Given a 3x3 matrix representing a unit cell, outputs a list of lattice
    parameters.

    Args:
        matrix: a 3x3 array or list, where the first, second, and third rows
            represent the a, b, and c vectors respectively
        radians: if True, outputs angles in radians. If False, outputs in
            degrees

    Returns:
        a 1x6 list of lattice parameters [a, b, c, alpha, beta, gamma]. a, b,
        and c are the length of the lattice vectos, and alpha, beta, and gamma
        are the angles between these vectors (in radians by default)
    """
    cell_para = np.zeros(6)
    # a
    cell_para[0] = np.linalg.norm(matrix[0])
    # b
    cell_para[1] = np.linalg.norm(matrix[1])
    # c
    cell_para[2] = np.linalg.norm(matrix[2])
    # alpha
    cell_para[3] = angle(matrix[1], matrix[2])
    # beta
    cell_para[4] = angle(matrix[0], matrix[2])
    # gamma
    cell_para[5] = angle(matrix[0], matrix[1])

    if not radians:
        # convert radians to degrees
        deg = 180.0 / np.pi
        cell_para[3] *= deg
        cell_para[4] *= deg
        cell_para[5] *= deg
    return cell_para


# def para2matrix(cell_para, radians=True, format="lower"):
def para2matrix(cell_para, radians=True, format="upper"):
    """
    Given a set of lattic parameters, generates a matrix representing the
    lattice vectors

    Args:
        cell_para: a 1x6 list of lattice parameters [a, b, c, alpha, beta,
            gamma]. a, b, and c are the length of the lattice vectos, and
            alpha, beta, and gamma are the angles between these vectors. Can
            be generated by matrix2para
        radians: if True, lattice parameters should be in radians. If False,
            lattice angles should be in degrees
        format: a string ('lower', 'symmetric', or 'upper') for the type of
            matrix to be output

    Returns:
        a 3x3 matrix representing the unit cell. By default (format='lower'),
        the a vector is aligined along the x-axis, and the b vector is in the
        y-z plane
    """
    a = cell_para[0]
    b = cell_para[1]
    c = cell_para[2]
    alpha = cell_para[3]
    beta = cell_para[4]
    gamma = cell_para[5]
    if radians is not True:
        alpha *= rad
        beta *= rad
        gamma *= rad
    cos_alpha = np.cos(alpha)
    cos_beta = np.cos(beta)
    cos_gamma = np.cos(gamma)
    sin_gamma = np.sin(gamma)
    sin_alpha = np.sin(alpha)
    matrix = np.zeros([3, 3])

    if format == "lower":
        # Generate a lower-diagonal matrix
        c1 = c * cos_beta
        c2 = (c * (cos_alpha - (cos_beta * cos_gamma))) / sin_gamma
        matrix[0][0] = a
        matrix[1][0] = b * cos_gamma
        matrix[1][1] = b * sin_gamma
        matrix[2][0] = c1
        matrix[2][1] = c2
        matrix[2][2] = np.sqrt(c**2 - c1**2 - c2**2)
    elif format == "symmetric":
        # TODO: allow generation of symmetric matrices
        pass
    elif format == "upper":
        # Generate an upper-diagonal matrix
        a3 = a * cos_beta
        a2 = (a * (cos_gamma - (cos_beta * cos_alpha))) / sin_alpha
        matrix[2][2] = c
        matrix[1][2] = b * cos_alpha
        matrix[1][1] = b * sin_alpha
        matrix[0][2] = a3
        matrix[0][1] = a2
        tmp = a**2 - a3**2 - a2**2
        if tmp > 0:
            matrix[0][0] = np.sqrt(a**2 - a3**2 - a2**2)
        # elif abs(tmp) < 1e-5: #tmp is very close to 0
        #    matrix[0][0] = 0
        #    print(matrix)
        else:
            # print(tmp)
            return None
        # pass
    return matrix


def gaussian_random_variable(min, max, sigma=3.0, random_state: None | int | Generator = None):
    """
    Choose a random number from a Gaussian probability distribution centered
    between min and max. sigma is the number of standard deviations that min
    and max are away from the center. Thus, sigma is also the largest possible
    number of standard deviations corresponding to the returned value. sigma=2
    corresponds to a 95.45% probability of choosing a number between min and
    max.

    Args:
        min: the minimum acceptable value
        max: the maximum acceptable value
        sigma: the number of standard deviations between the center and min/max

    Returns:
        a value chosen randomly between min and max
    """
    if isinstance(random_state, Generator):
        random_state = random_state.spawn(1)[0]
    else:
        random_state = np.random.default_rng(random_state)

    center = (max + min) * 0.5
    delta = np.fabs(max - min) * 0.5
    ratio = delta / sigma
    while True:
        x = random_state.normal(scale=ratio, loc=center)
        if x > min and x < max:
            return x


def random_vector(minvec=None, maxvec=None, width=0.35, unit=False, random_state: None | int | Generator = None):
    """
    Generate a random vector for lattice constant generation. The ratios between
    x, y, and z of the returned vector correspond to the ratios between a, b,
    and c. Results in a Gaussian distribution of the natural log of the ratios.

    Args:
        minvec: the bottom-left-back minimum point which can be chosen
        maxvec: the top-right-front maximum point which can be chosen
        width: the width of the normal distribution to use when choosing values.
            Passed to np.random.normal
        unit: whether or not to normalize the vector to determinant 1

    Returns:
        a 1x3 numpy array of floats
    """
    if isinstance(random_state, Generator):
        random_state = random_state.spawn(1)[0]
    else:
        random_state = np.random.default_rng(random_state)

    # TODO these were not used in the original code, moved out of the defaults
    # and declared if None to avoid having lists as default arguments.
    if maxvec is None:
        maxvec = [1.0, 1.0, 1.0]
    if minvec is None:
        minvec = [0.0, 0.0, 0.0]

    vec = np.exp(random_state.normal(scale=width, size=3))
    if unit:
        return vec / np.linalg.norm(vec)
    else:
        return vec


def random_shear_matrix(width=1.0, unitary=False, random_state: None | int | Generator = None):
    """
    Generate a random symmetric shear matrix with Gaussian elements. If unitary
    is True, normalize to determinant 1

    Args:
        width: the width of the normal distribution to use when choosing values.
            Passed to np.random.normal
        unitary: whether or not to normalize the matrix to determinant 1

    Returns:
        a 3x3 numpy array of floats
    """
    if isinstance(random_state, Generator):
        random_state = random_state.spawn(1)[0]
    else:
        random_state = np.random.default_rng(random_state)

    mat = np.zeros([3, 3])
    determinant = 0
    while determinant == 0:
        a, b, c = random_state.normal(scale=width, size=3)
        mat = np.array([[1, a, b], [a, 1, c], [b, c, 1]])
        determinant = np.linalg.det(mat)
    if unitary:
        return mat / np.cbrt(np.linalg.det(mat))
    else:
        return mat



import os
from time import time

import numpy as np
import torchani
from ase.constraints import UnitCellFilter, FixSymmetry
from ase.optimize.fire import FIRE
#from ase.spacegroup.symmetrize import FixSymmetry


def ANI_relax(struc, opt_cell=False, step=500, fmax=0.1, logfile=None, max_time=6.0):
    """
    ani optimizer
    Args:
    struc: ase atoms object
    step: optimization steps (int)
    max_time: float (minutes)
    """
    calc = torchani.models.ANI2x().ase()
    struc.set_calculator(calc)
    struc.set_constraint(FixSymmetry(struc))
    if opt_cell:
        ecf = UnitCellFilter(struc)
        dyn = FIRE(ecf, a=0.1, logfile=logfile) if logfile is not None else FIRE(ecf, a=0.1)
    else:
        dyn = FIRE(struc, a=0.1, logfile=logfile) if logfile is not None else FIRE(struc, a=0.1)

    # Run relaxation
    if step < 50:
        dyn.run(fmax=fmax, steps=step)
    else:
        t0 = time()
        dyn.run(fmax=fmax, steps=int(step / 2))
        # If time is too long, only run half steps
        if (time() - t0) / 60 < max_time / 2:
            dyn.run(fmax=fmax, steps=int(step / 2))
    return struc


class ANI:
    """
    This is a calculator to perform oragnic crystal structure optimization in ANI
    We assume that the geometry has been well optimized by classical FF

    Args:
        struc: pyxtal object
        opt: 'conv', 'conp', 'single'
    """

    def __init__(self, struc, opt_lat=True, logfile=None):
        self.structure = struc
        self.opt_lat = opt_lat
        self.stress = None
        self.forces = None
        self.optimized = True
        self.positions = None
        self.cell = None
        self.cputime = 0
        self.logfile = logfile

    def run(self, steps=10):
        t0 = time()
        s = self.structure.to_ase(resort=False)
        s.set_constraint(FixSymmetry(s))
        s.set_calculator(torchani.models.ANI2x().ase())#; print("Setup Fire")

        if not self.opt_lat:
            dyn = FIRE(s, a=0.1, logfile=self.logfile)#, force_consistent=False)
            dyn.run(fmax=0.1, steps=steps)
        else:
            #ecf = FrechetCellFilter(s)
            ecf = UnitCellFilter(s)
            dyn = FIRE(ecf, a=0.1, logfile=self.logfile)#, force_consistent=False)
            dyn.run(fmax=0.1, steps=steps)
            self.structure.lattice.set_matrix(s.get_cell())

        positions = s.get_scaled_positions()
        #try:
        if True:
            # s.write('../1.cif', format='cif')
            count = 0
            for _i, site in enumerate(self.structure.mol_sites):
                coords0, _ = site._get_coords_and_species(first=True)
                coords1 = positions[count : count + len(site.molecule.mol)]
                for j, coor in enumerate(coords1):
                    diff = coor - coords0[j]
                    diff -= np.round(diff)
                    abs_diff = np.dot(diff, s.get_cell())
                    # print(j, coor, coords0[j], diff, np.linalg.norm(abs_diff))
                    if abs(np.linalg.norm(abs_diff)) < 2.0:
                        coords1[j] = coords0[j] + diff
                    else:
                        print(coords1[j], coords1[j], np.linalg.norm(abs_diff))
                        import sys; sys.exit()

                site.update(coords1, self.structure.lattice)
                count += len(site.molecule.mol) * site.wp.multiplicity
            self.structure.optimize_lattice()
            self.structure.energy = s.get_potential_energy()
            self.cell = s.get_cell()
            # print(self.structure.lattice)
        #except:
        #    self.structure.energy = 10000
        #    self.optimized = False
        #    print("Structure is wrong after optimization")
        s.set_calculator()
        s.set_constraint()
        self.cputime = time() - t0


if __name__ == "__main__":
    import warnings

    from pyxtal.db import database

    warnings.filterwarnings("ignore")

    work_dir = "tmp"
    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    db = database("benchmarks/test.db")
    struc = db.get_pyxtal("ACSALA")

    calc = ANI(struc)
    print(calc.structure.lattice)
    calc.run()
    print(calc.structure.energy)
    print(calc.structure.lattice)



"""
Module for generating molecular crystals
"""

# Standard Libraries
from copy import deepcopy

import numpy as np

from pyxtal.lattice import Lattice
from pyxtal.molecule import pyxtal_molecule
from pyxtal.msg import Comp_CompatibilityError, Symm_CompatibilityError, VolumeError
from pyxtal.symmetry import Group
from pyxtal.symmetry import choose_wyckoff_mol as wyc_mol
from pyxtal.tolerance import Tol_matrix
from pyxtal.wyckoff_site import mol_site


# Define functions
# ------------------------------
class molecular_crystal:
    """
    Class for storing and generating molecular crystals based on symmetry
    constraints. Based on the crystal.random_crystal class for atomic crystals.
    Given a spacegroup, list of molecule objects, molecular stoichiometry, and
    a volume factor, generates a molecular crystal consistent with the given
    constraints.

    Args:
        dim: dimenion (1, 2, 3)
        group: the group number (1-75, 1-80, 1-230)
        molecules: a list of pymatgen.core.structure.Molecule objects for
            each type of molecule. Alternatively, you may supply a file path,
            or the name of molecules from the built_in
            `database <pyxtal.database.collection.html>`_
        numMols: A list of the number of each type of molecule within the
            primitive cell (NOT the conventioal cell)
        factor: A volume factor used to generate a larger or smaller
            unit cell. Increasing this gives extra space between molecules
        lattice (optional): the `Lattice <pyxtal.lattice.Lattice.html>`_
            object to define the unit cell
        conventional (optional): count the atomic numbers in a conventional cell
        tm (optional): the `Tol_matrix <pyxtal.tolerance.tolerance.html>`_
            object to define the distances
        sites (optional): pre-assigned wyckoff sites (e.g., `[["4a"], ["2b"]]`)
        seed (optional): seeds
        use_hall: False
    """

    def __init__(
        self,
        dim,
        group,
        molecules,
        numMols,
        factor=1.1,
        thickness=None,
        area=None,
        lattice=None,
        torsions=None,
        tm=Tol_matrix(prototype="molecular"),
        sites=None,
        conventional=True,
        seed=None,
        random_state=None,
        use_hall=False,
    ):
        # Initialize
        self.source = "Random"
        self.valid = False
        self.factor = factor
        self.seed = seed
        self.random_state = np.random.default_rng(random_state)

        # Dimesion
        self.dim = dim
        self.area = area  # Cross-section area for 1D
        self.thickness = thickness  # Thickness of 2D slab

        # The periodic boundary condition
        if dim == 3:
            self.PBC = [1, 1, 1]
        elif dim == 2:
            self.PBC = [1, 1, 0]
        elif dim == 1:
            self.PBC = [0, 0, 1]

        # Symmetry Group
        if type(group) == Group:
            self.group = group
        else:
            self.group = Group(group, dim=self.dim, use_hall=use_hall)
        self.number = self.group.number
        self.hall_number = self.group.hall_number

        # Composition
        if numMols is None:
            numMols = [len(self.group[0])] * len(molecules)
            no_check_compability = True
        else:
            numMols = np.array(numMols)  # must convert it to np.array
            no_check_compability = False
        mul = self.group.cellsize() if not conventional else 1
        self.numMols = numMols * mul

        # Tolerance matrix
        if type(tm) == Tol_matrix:
            self.tol_matrix = tm
        else:
            self.tol_matrix = Tol_matrix(prototype=tm)

        # Wyckofff sites
        self.set_molecules(molecules, torsions)
        self.set_sites(sites)
        valid_orientation = self.set_orientations()

        if not valid_orientation:
            msg = "Molecular symmetry is compatible with WP site\n"
            msg += "".join(f"{mol}: {mol.pga.sch_symbol}, " for mol in self.molecules)
            raise Symm_CompatibilityError(msg)

        # Check the minimum dof within the Wyckoff positions
        if no_check_compability:
            compat, self.degrees = True, True
        else:
            compat, self.degrees = self.group.check_compatible(self.numMols, self.valid_orientations)

        if not compat:
            msg = f"Compoisition {self.numMols} not compatible with symmetry {self.group.number}"
            raise Comp_CompatibilityError(msg)

        self.set_volume()
        self.set_lattice(lattice)
        self.set_crystal()

    def __str__(self):
        s = "------Random Molecular Crystal------"
        s += f"\nDimension: {self.dim}"
        # s += f"\nGroup: {self.symbol}"
        s += f"\nVolume factor: {self.factor}"
        s += f"\n{self.lattice}"
        if self.valid:
            s += "\nWyckoff sites:"
            s += "".join(f"\n\t{wyc}" for wyc in self.mol_sites)
        else:
            s += "\nStructure not generated."
        return s

    def __repr__(self):
        return str(self)

    def set_sites(self, sites):
        """
        initialize Wyckoff sites

        Args:
            sites: list
        """
        # Symmetry sites
        self.sites = {}
        for i, _mol in enumerate(self.molecules):
            if sites is not None and sites[i] is not None and len(sites[i]) > 0:
                self._check_consistency(sites[i], self.numMols[i])
                if isinstance(sites[i], dict):
                    self.sites[i] = []
                    for item in sites[i].items():
                        self.sites[i].append({item[0]: item[1]})
                else:
                    self.sites[i] = sites[i]
            else:
                self.sites[i] = None

    def set_molecules(self, molecules, torsions):
        """
        Get molecular information

        Args:
            molecules: list of molecules
            torsions: list of torsions
        """
        if torsions is None:
            torsions = [None] * len(molecules)

        self.molecules = []
        for i, mol in enumerate(molecules):
            # already a pyxtal_molecule object
            if isinstance(mol, pyxtal_molecule):
                p_mol = mol
            else:
                p_mol = pyxtal_molecule(mol, seed=self.seed, torsions=torsions[i], tm=self.tol_matrix, random_state=self.random_state)
            self.molecules.append(p_mol)

    def set_orientations(self):
        """
        Calculates the valid orientations for each Molecule and Wyckoff
        position. Returns a list with 4 indices:
            - index 1: the molecular prototype's index within self.molecules
            - index 2: the WP's 1st index (based on multiplicity)
            - index 3: the WP's 2nd index (within the group of same multiplicity)
            - index 4: the index of a valid orientation for the molecule/WP pair

        For example, self.valid_orientations[i][j][k] would be a list of valid
        orientations for self.molecules[i], in the Wyckoff position
        self.group.wyckoffs_organized[j][k]
        """
        valid_ori = False
        self.valid_orientations = []
        for i, pyxtal_mol in enumerate(self.molecules):
            self.valid_orientations.append([])
            for x in self.group.wyckoffs_organized:
                self.valid_orientations[-1].append([])
                for _j, wp in enumerate(x):
                    # Don't check the wp with high multiplicity
                    if len(wp) > self.numMols[i]:
                        allowed = []
                    else:
                        allowed = pyxtal_mol.get_orientations_in_wp(wp)
                        if len(allowed) > 0:
                            valid_ori = True
                    self.valid_orientations[-1][-1].append(allowed)
        return valid_ori

    def set_volume(self):
        """
        Given the molecular stoichiometry, estimate the volume for a unit cell.
        """
        volume = 0
        for numMol, mol in zip(self.numMols, self.molecules):
            volume += numMol * mol.volume
        self.volume = abs(self.factor * volume)

    def set_lattice(self, lattice):
        """
        Generate the initial lattice
        """
        if lattice is not None:
            # Use the provided lattice
            self.lattice = lattice
            self.volume = lattice.volume
            # Make sure the custom lattice PBC axes are correct.
            if lattice.PBC != self.PBC:
                self.lattice.PBC = self.PBC
                raise ValueError("PBC is incompatible " + str(self.PBC))
        else:
            # Determine the unique axis
            if self.dim == 2:
                unique_axis = "c" if self.number in range(3, 8) else "a"
            elif self.dim == 1:
                unique_axis = "a" if self.number in range(3, 8) else "c"
            else:
                unique_axis = "c"

            # Generate a Lattice instance
            good_lattice = False
            for _cycle in range(10):
                try:
                    if self.group.number < 10:
                        coef = 1.0 * self.numMols[0] / self.group[0].multiplicity
                    elif 10 <= self.group.number <= 15:
                        coef = 2.0  # 2/m
                    elif 16 <= self.group.number <= 74:
                        coef = 1.5
                    else:
                        coef = 1.0

                    self.lattice = Lattice(
                        self.group.lattice_type,
                        self.volume,
                        PBC=self.PBC,
                        unique_axis=unique_axis,
                        thickness=self.thickness,
                        area=self.area,
                        min_special=coef * max([mol.get_max_length() for mol in self.molecules]),
                        random_state=self.random_state,
                    )
                    good_lattice = True
                    break
                except VolumeError:
                    self.volume *= 1.1
                    msg = "Warning: increase the volume by 1.1 times: "
                    msg += f"{self.volume:.2f}"
                    print(msg)

            if not good_lattice:
                msg = f"Volume estimation {self.volume:.2f} is very bad"
                msg += " with the given composition "
                msg += str(self.numMols)
                raise RuntimeError(msg)

    def set_crystal(self):
        """
        The main code to generate a random molecular crystal.
        If successful, `self.valid` is True
        """
        self.numattempts = 0
        if not self.degrees:
            self.lattice_attempts = 20
            self.coord_attempts = 3
            self.ori_attempts = 1
        else:
            self.lattice_attempts = 40
            self.coord_attempts = 30
            self.ori_attempts = 5

        if not self.lattice.allow_volume_reset:
            self.lattice_attempts = 1

        for cycle1 in range(self.lattice_attempts):
            self.cycle1 = cycle1
            for cycle2 in range(self.coord_attempts):
                self.cycle2 = cycle2
                output = self._set_coords()

                if output:
                    self.mol_sites = output
                    break
            if self.valid:
                return
            else:
                self.lattice.reset_matrix()

        print("Cannot generate crystal after max attempts.")

    def _set_coords(self):
        """
        generate coordinates for random crystal
        """

        mol_sites_total = []
        # Add molecules
        for i, numMol in enumerate(self.numMols):
            pyxtal_mol = self.molecules[i]
            valid_ori = self.valid_orientations[i]
            output = self._set_mol_wyckoffs(i, numMol, pyxtal_mol, valid_ori, mol_sites_total)
            if output is not None:
                mol_sites_total.extend(output)
            else:
                # correct multiplicity not achieved exit and start over
                return None

        self.valid = True
        return mol_sites_total

    def _set_mol_wyckoffs(self, id, numMol, pyxtal_mol, valid_ori, mol_wyks):
        """
        generates a set of wyckoff positions to accomodate a given number
        of molecules

        Args:
            id: molecular id
            numMol: Number of ions to accomodate
            pyxtal_mol: Type of species being placed on wyckoff site
            valid_ori: list of valid orientations
            mol_wyks: current wyckoff sites

        Returns:
            if sucess, wyckoff_sites_tmp: list of wyckoff sites for valid sites
            otherwise, None

        """
        numMol_added = 0
        mol_sites_tmp = []

        # Now we start to add the specie to the wyckoff position
        sites_list = deepcopy(self.sites[id])  # the list of Wyckoff site
        if sites_list is not None:
            self.wyckoff_attempts = max(len(sites_list) * 2, 10)
        else:
            # the minimum numattempts is to put all atoms to the general WPs
            min_wyckoffs = int(numMol / len(self.group.wyckoffs_organized[0][0]))
            self.wyckoff_attempts = max(2 * min_wyckoffs, 10)

        for _cycle in range(self.wyckoff_attempts):
            # Choose a random WP for given multiplicity: 2a, 2b, 2c
            if sites_list is not None and len(sites_list) > 0:
                site = sites_list[0]
            else:  # Selecting the merging
                site = None

            # NOTE: The molecular version return wyckoff indices, not ops
            diff = numMol - numMol_added

            if type(site) is dict:  # site with coordinates
                key = next(iter(site.keys()))
                wp = wyc_mol(self.group, diff, key, valid_ori, True, self.dim, self.random_state)
            else:
                wp = wyc_mol(self.group, diff, site, valid_ori, True, self.dim, self.random_state)

            if wp is not False:
                # Generate a list of coords from the wyckoff position
                mult = wp.multiplicity  # remember the original multiplicity

                pt = site[key] if type(site) is dict else self.lattice.generate_point()
                # merge coordinates if the atoms are close
                mtol = pyxtal_mol.radius * 0.5
                pt, wp, oris = wp.merge(pt, self.lattice.matrix, mtol, valid_ori, self.group)

                if wp is not False:
                    if site is not None and mult != wp.multiplicity:
                        continue
                    if self.dim == 2 and self.thickness is not None and self.thickness < 0.1:
                        pt[-1] = 0.5

                    ms0 = self._set_orientation(pyxtal_mol, pt, oris, wp)
                    if ms0 is not None:
                        # Check current WP against existing WP's
                        passed_wp_check = True
                        # print("Checking", ms0)
                        for ms1 in mol_wyks + mol_sites_tmp:
                            if not ms0.short_dist_with_wp2(ms1, tm=self.tol_matrix):
                                passed_wp_check = False
                            # else:
                            #    print("passing", ms1)

                        if passed_wp_check:
                            if sites_list is not None:
                                sites_list.pop(0)

                            ms0.type = id
                            mol_sites_tmp.append(ms0)
                            numMol_added += len(ms0.wp)

                            # We have enough molecules of the current type
                            if numMol_added == numMol:
                                return mol_sites_tmp
        return None

    def _set_orientation(self, pyxtal_mol, pt, oris, wp):
        """
        Generate good orientations
        """
        # Use a Wyckoff_site object for the current site
        self.numattempts += 1
        # NOTE removing this copy causes tests to fail -> state not managed well
        ori = self.random_state.choice(oris).copy()
        ori.change_orientation(flip=True)
        ms0 = mol_site(pyxtal_mol, pt, ori, wp, self.lattice)
        # Check distances within the WP
        if ms0.short_dist():
            return ms0
        else:
            # Maximize the smallest distance for the general
            # positions if needed
            if len(pyxtal_mol.mol) > 1 and ori.degrees > 0:
                # bisection method
                def fun_dist(angle, ori, mo, pt):
                    # ori0 = ori.copy()
                    ori.change_orientation(angle)
                    ms0 = mol_site(
                        mo,
                        pt,
                        ori,
                        wp,
                        self.lattice,
                    )
                    return ms0.get_min_dist()

                angle_lo = ori.angle
                angle_hi = angle_lo + np.pi
                fun_lo = fun_dist(angle_lo, ori, pyxtal_mol, pt)
                fun_hi = fun_dist(angle_hi, ori, pyxtal_mol, pt)
                fun = fun_hi
                for _it in range(self.ori_attempts):
                    self.numattempts += 1
                    if (fun > 0.8) & (ms0.short_dist()):
                        return ms0
                    angle = (angle_lo + angle_hi) / 2
                    fun = fun_dist(angle, ori, pyxtal_mol, pt)
                    # print('Bisection: ', it, fun)
                    if fun_lo > fun_hi:
                        angle_hi, fun_hi = angle, fun
                    else:
                        angle_lo, fun_lo = angle, fun

        return None

    def _check_consistency(self, site, numMol):
        """
        Check if the composition is consistent with symmetry
        """
        num = 0
        for s in site:
            num += int(s[:-1])
        if numMol == num:
            return True
        else:
            msg = "\nThe requested number of molecules is inconsistent: "
            msg += str(site)
            msg += f"\nfrom numMols: {numMol:d}"
            msg += f"\nfrom Wyckoff list: {num:d}"
            raise ValueError(msg)



import contextlib
import os
import shutil

import numpy as np


class CHARMM:
    """
    This is a calculator to perform oragnic crystal structure optimization in GULP
    Args:
        struc: pyxtal.molecular_crystal.molecular_crystal object
        rtf: path of rtf file
        prm: path of prm file
        opt: 'conv', 'conp', 'single'
        input: charmm input file
        output: charmm output file
        dump: charmm dump structure
        exe: charmm executable
    """

    def __init__(
        self,
        struc,
        label="_",
        algo="abnr",
        lat_mut=False,
        rotate=False,
        prefix="pyxtal",
        atom_info=None,
        folder=".",
        opt="conp",
        steps=None,
        exe="charmm",
        input="charmm.in",
        output="charmm.log",
        dump="result.pdb",
        debug=False,
    ):
        if steps is None:
            steps = [2000, 1000]
        self.debug = debug

        # check charmm Executable
        if shutil.which(exe) is None:
            raise BaseException(f"{exe} is not installed")
        else:
            self.exe = exe

        self.atom_info = atom_info

        # Optimization setting
        self.opt = opt
        self.optlat = False
        self.algo = algo
        if type(steps) is list:
            self.steps = steps
        elif type(steps) is int:
            self.steps = [2000, steps]
        self.opt = opt

        # Files IO
        self.folder = folder
        self.prefix = prefix
        self.label = label
        self.rtf = self.prefix + ".rtf"
        self.prm = self.prefix + ".prm"
        self.psf = self.label + self.prefix + ".psf"
        self.crd = self.label + self.prefix + ".crd"
        self.input = self.label + input
        self.output = self.label + output
        self.dump = self.label.lower() + dump  # charmm only output lower

        # For parsing the output
        self.positions = None
        self.forces = None
        self.optimized = False
        self.cell = None
        self.cputime = 0.0

        # Structure Manipulation
        struc.resort()
        self.structure = struc
        try:
            self.structure.optimize_lattice()
        except:
            print("bug")
            print(self.structure)
            print(self.structure.lattice)
            print(self.structure.lattice.matrix)
            raise ValueError("Problem in Lattice")
        # print("\nbeginining lattice: ", struc.lattice)

        self.lat_mut = lat_mut
        self.rotate = rotate

    def run(self, clean=True):
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        cwd = os.getcwd()
        os.chdir(self.folder)

        self.write()  # ; print("write", time()-t0)
        self.execute()  # ; print("exe", time()-t0)
        self.read()  # ; print("read", self.structure.energy)
        if clean:
            self.clean()

        os.chdir(cwd)

    def execute(self):
        cmd = self.exe + "<" + self.input + ">" + self.output
        os.system(cmd)

    def clean(self):
        os.remove(self.input)
        os.remove(self.output)
        os.remove(self.crd) if os.path.exists(self.crd) else None
        os.remove(self.psf) if os.path.exists(self.psf) else None
        os.remove(self.dump) if os.path.exists(self.dump) else None

    def write(self):
        """
        setup the necessary files for charmm calculation
        """
        lat = self.structure.lattice
        if self.lat_mut:
            lat = lat.mutate()

        a, b, c, alpha, beta, gamma = lat.get_para(degree=True)
        ltype = lat.ltype
        if ltype in ['trigonal', 'Trigonal']: ltype = 'hexagonal'

        fft = self.FFTGrid(np.array([a, b, c]))

        with open(self.input, "w") as f:
            # General input
            f.write("! Automated Charmm calculation\n\n")
            f.write("bomlev -1\n")
            f.write("! top and par\n")
            f.write(f"read rtf card name {self.rtf:s}\n")
            f.write(f"read para card name {self.prm:s}\n")
            f.write("Read sequence card\n")
            f.write(f"{len(self.structure.mol_sites):5d}\n")
            atom_count = []
            for site in self.structure.mol_sites:
                atom_count.append(len(site.molecule.mol))
                if self.atom_info is None:
                    f.write(f"U0{site.type:d} ")
                else:
                    f.write("{:s} ".format(self.atom_info["resName"][site.type]))

            f.write("\ngenerate main first none last none setup warn\n")
            f.write("Read coor card free\n")
            f.write("* Residues coordinate\n*\n")
            f.write(f"{sum(atom_count):5d}\n")
            for i, site in enumerate(self.structure.mol_sites):
                res_name = f"U0{site.type:d}" if self.atom_info is None else self.atom_info["resName"][site.type]

                # reset lattice if needed (to move out later)
                site.lattice = lat

                # Rotate if needed
                if self.rotate:
                    site.perturbate(lat.matrix, trans=0.5, rot="random")
                coords, species = site._get_coords_and_species(first=True)
                if i == 0:
                    count = 0
                else:
                    count += atom_count[i - 1]

                for j, coord in enumerate(coords):
                    if self.atom_info is None:
                        label = f"{species[j]:s}{j + 1:d}"
                    else:
                        label = self.atom_info["label"][site.type][j]
                    f.write(
                        "{:5d}{:5d}{:>4s}  {:<4s}{:10.5f}{:10.5f}{:10.5f}\n".format(
                            j + 1 + count, i + 1, res_name, label, *coord
                        )
                    )

            f.write(f"write psf card name {self.psf:s}\n")
            f.write(f"write coor crd card name {self.crd:s}\n")
            f.write(f"read psf card name {self.psf:s}\n")
            f.write(f"read coor card name {self.crd:s}\n")

            # Structure info
            f.write("\n! crystal parameters\n")
            f.write(f"set shape {ltype:s}\n")
            f.write(f"set a     {a:12.6f}\n")
            f.write(f"set b     {b:12.6f}\n")
            f.write(f"set c     {c:12.6f}\n")
            f.write(f"set alpha {alpha:12.6f}\n")
            f.write(f"set beta  {beta:12.6f}\n")
            f.write(f"set gamma {gamma:12.6f}\n")
            f.write("coor conv FRAC SYMM @a @b @c @alpha @beta @gamma\n")
            f.write("coor stat select all end\n")
            f.write("Crystal Define @shape @a @b @c @alpha @beta @gamma\n")
            site0 = self.structure.mol_sites[0]
            f.write(f"Crystal Build cutoff 14.0 noperations {len(site0.wp.ops) - 1:d}\n")
            for i, op in enumerate(site0.wp.ops):
                if i > 0:
                    f.write(f"({op.as_xyz_str():s})\n")

            f.write("image byres xcen ?xave ycen ?yave zcen ?zave sele resn LIG end\n")
            f.write("set 7 fswitch\n")
            f.write("set 8 atom\n")
            f.write("set 9 vatom\n")
            f.write("Update inbfrq 10 imgfrq 10 ihbfrq 10 -\n")
            f.write("ewald pmewald lrc fftx {:d} ffty {:d} fftz {:d} -\n".format(*fft))
            f.write("kappa 0.34 order 6 CTOFNB 12.0 CUTNB 14.0 QCOR 1.0 -\n")
            f.write("@7 @8 @9 vfswitch !\n")
            f.write(f"mini {self.algo:s} nstep {self.steps[0]:d}\n")
            if len(self.steps) > 1:
                f.write(f"mini {self.algo:s} lattice nstep {self.steps[1]:d} \n")
            if len(self.steps) > 2:
                f.write(f"mini {self.algo:s} nstep {self.steps[2]:d}\n")

            f.write("coor conv SYMM FRAC ?xtla ?xtlb ?xtlc ?xtlalpha ?xtlbeta ?xtlgamma\n")  #
            f.write(f"\nwrite coor pdb name {self.dump:s}\n")  #
            f.write("*CELL :  ?xtla  ?xtlb  ?xtlc ?xtlalpha ?xtlbeta ?xtlgamma\n")  #
            f.write(f"*Z = {len(site0.wp):d}\n")
            f.write("*Energy(kcal): ?ener\n")
            f.write("stop\n")
        # print("STOP")
        # import sys
        # sys.exit()

    def read(self):
        with open(self.output) as f:
            lines = f.readlines()
            self.version = lines[2]
            if lines[-1].find("CPU TIME") != -1:
                self.optimized = True
                self.cputime = float(lines[-2].split()[-2])
                for line in lines:
                    if line.find("MINI> ") != -1:
                        with contextlib.suppress(Exception):
                            self.structure.iter = int(line.split()[1])
                            # raise RuntimeError("Something is wrong in the charmm output")
                    elif line.find("ABNORMAL TERMINATION") != -1:
                        self.optimized = False
                        break

        if self.optimized:
            with open(self.dump) as f:
                lines = f.readlines()
                positions = []
                for line in lines:
                    if line.find("REMARK ENERGY") != -1:
                        tmp = line.split(":")
                        self.structure.energy = float(tmp[-1])
                    elif line.find("REMARK Z") != -1:
                        tmp = line.split(":")[-1].split()
                        Z = int(tmp[-1])
                    elif line.find("REMARK CELL") != -1:
                        tmp = line.split(":")[-1].split()
                        tmp = [float(x) for x in tmp]
                        self.structure.lattice.set_para(tmp)
                    elif line.find("ATOM") != -1:
                        try:
                            xyz = line.split()[5:8]
                            XYZ = [float(x) for x in xyz]
                            positions.append(XYZ)
                        except:
                            pass  # print("Warning: BAD charmm output: " + line)
                positions = np.array(positions)
                self.structure.energy *= Z

            count = 0
            # for i, site in enumerate(self.structure.mol_sites):
            #    coords = positions[count:count+len(site.mol)]
            #    site.update(coords, self.structure.lattice)
            #    count += len(site.mol)

            # if True:
            try:
                for _i, site in enumerate(self.structure.mol_sites):
                    coords = positions[count : count + len(site.molecule.mol)]
                    site.update(coords, self.structure.lattice)
                    count += len(site.molecule.mol)
                # print("after relaxation  : ", self.structure.lattice, "iter: ", self.structure.iter)
                self.structure.optimize_lattice()
                # print("after latticeopt  : ", self.structure.lattice, self.structure.check_distance()); import sys; sys.exit()
            except:
                # molecular connectivity or lattice optimization
                self.structure.energy = 10000
                if self.debug:
                    print("Unable to retrieve Structure after optimization")
                    print("lattice", self.structure.lattice)
                    self.structure.to_file("1.cif")
                    print("Check 1.cif in ", os.getcwd())
                    pairs = self.structure.check_short_distances()
                    if len(pairs) > 0:
                        print(self.structure.to_file())
                        print("short distance pair", pairs)

        else:
            self.structure.energy = 10000
            if self.debug:
                print(self.structure)
                import sys

                sys.exit()

    def FFTGrid(self, ABC):
        """
        determine the grid used for
        """

        grid = np.array(
            [
                1,
                2,
                3,
                4,
                5,
                6,
                8,
                9,
                10,
                12,
                15,
                16,
                18,
                20,
                24,
                25,
                27,
                30,
                32,
                36,
                40,
                45,
                48,
                50,
                54,
                60,
                72,
                75,
                80,
                81,
                90,
                96,
                100,
                108,
                120,
                125,
                135,
                144,
                150,
                160,
                162,
                180,
                200,
                216,
                225,
                240,
                243,
                250,
                270,
                288,
                300,
                324,
                360,
                375,
                400,
            ]
        )
        fftxyz = [0, 0, 0]
        for i, l in enumerate(ABC):
            tmp = grid[grid > l]
            fftxyz[i] = max([32, tmp[0]])  # minimum is 32
        return fftxyz


def check_prm(path):
    """
    Todo: move it to charmm interface
    Sometimes there are something wrong with prm file (RUBGIH),
    """
    with open(path) as f:
        lines = f.readlines()
        pairs = []
        triplets = []
        imphis = []
        ids = []
        do_angle = False
        do_bond = False
        do_imphi = False
        for i, l in enumerate(lines):
            tmp = l.split()
            if l.find("ANGLE") == 0:
                do_angle = True
            elif l.find("BOND") == 0:
                do_bond = True
            elif l.find("DIHEDRAL") == 0:
                pass
            elif l.find("IMPHI") == 0:
                do_imphi = True
            elif do_bond:
                if len(tmp) > 0:
                    pair1 = tmp[0] + " " + tmp[1]
                    pair2 = tmp[1] + " " + tmp[0]
                    if (pair1 in pairs) or (pair2 in pairs):
                        ids.append(i)
                        print("Duplicate bonds: ", l[:-2])
                    else:
                        pairs.append(pair1)
                else:
                    do_bond = False
            elif do_angle:
                if len(tmp) > 0:
                    pair1 = tmp[0] + " " + tmp[1] + " " + tmp[2]
                    pair2 = tmp[2] + " " + tmp[1] + " " + tmp[0]
                    if (pair1 in triplets) or (pair2 in triplets):
                        ids.append(i)
                        print("Duplicate angles: ", l[:-2])
                    else:
                        triplets.append(pair1)
                else:
                    do_angle = False
            # elif do_dihedral:
            #    if len(tmp) > 0:
            #        pair1 = tmp[0]+ ' ' + tmp[1] + ' ' + tmp[2] + ' ' +tmp[3] + ' ' + tmp[5]
            #        pair2 = tmp[3]+ ' ' + tmp[1] + ' ' + tmp[2] + ' ' +tmp[0] + ' ' + tmp[5]
            #        if (pair1 in quads) or (pair2 in quads):
            #            ids.append(i)
            #            #print(pair1)
            #            #print(pair2)
            #            #print(quads)
            #            print('Duplicate dihedral angles: ', l[:-2])
            #            #import sys
            #            #sys.exit()
            #        else:
            #            quads.append(pair1)
            #    else:
            #        do_dihedral = False
            elif do_imphi:
                if len(tmp) > 0:
                    pair1 = tmp[0] + " " + tmp[1] + " " + tmp[2] + " " + tmp[3]
                    # pair2 = tmp[0] + ' ' + tmp[1] + ' ' + tmp[2] + ' ' +tmp[0]
                    if pair1 in imphis:  # or (pair2 in imphis):
                        ids.append(i)
                        print("Duplicate imphi angles: ", l[:-2])
                    else:
                        imphis.append(pair1)
                else:
                    do_imphi = False
                    break

    lines = [lines[i] for i in range(len(lines)) if i not in ids]
    with open(path, "w") as f:
        f.writelines(lines)


class RTF:
    """
    Parse, convert, merge the RTF files for CHARMM
    """

    def __init__(self, input):
        self.keywords = [
            "MASS",
            "RESI",
            "CHARGE",
            "ATOM",
            "BOND",
            "ANGL",
            "DIHE",
            "IMPH",
        ]
        if type(input) == str:
            print("Converting RTF from the file", input)
            f = open(input)
            rtf = f.read().split("\n")
            f.close()
        else:
            rtf = input

        self.parse(rtf)

    def parse(self, rtf):
        """
        parse rtf list to get the dictionary
        """
        mass = []
        self.residues = []

        do_resi = False
        residue = None
        for l in rtf:
            tmp = l.split()
            if l.find("MASS") == 0:
                mass.append(l)
                do_resi = False
            elif l.find("RESI") == 0:
                if residue is not None:
                    self.residues.append(residue)
                do_resi = True
                residue = {
                    "NAME": tmp[1],
                    "CHARGE": float(tmp[2]),
                    "ATOM": [],
                    "BOND": [],
                    "ANGL": [],
                    "DIHE": [],
                    "IMPH": [],
                }

            if do_resi:
                if l.find("ATOM") == 0:  # len(tmp) > 0:
                    residue["ATOM"].append(l)
                elif l.find("BOND") == 0:  # len(tmp) > 0:
                    residue["BOND"].append(l)
                elif l.find("ANGL") == 0:  # len(tmp) > 0:
                    residue["ANGL"].append(l)
                elif l.find("DIHE") == 0:  # len(tmp) > 0:
                    residue["DIHE"].append(l)
                elif l.find("IMPH") == 0:  # len(tmp) > 0:
                    residue["IMPH"].append(l)
        self.parse_mass(mass)
        self.residues.append(residue)

    def to_string(self):
        """
        convert the RTF object to string
        """

        strs = "* Topology File.\n"
        strs += "*\n   99   1\n"
        for l, m in zip(self.labels, self.mass):
            strs += f"MASS   -1 {l:4s} {m:12.6f}\n"

        for res in self.residues:
            # print(res)
            strs += "\nRESI {:3s} {:5.3f}\n".format(res["NAME"], res["CHARGE"])
            strs += "GROUP\n"
            for a in res["ATOM"]:
                strs += f"{a:s}\n"
            strs += "\n"
            for b in res["BOND"]:
                strs += f"{b:s}\n"
            strs += "\n"
            for a in res["ANGL"]:
                strs += f"{a:s}\n"
            strs += "\n"
            for d in res["DIHE"]:
                strs += f"{d:s}\n"
            strs += "\n"
            for i in res["IMPH"]:
                strs += f"{i:s}\n"

        return strs

    def to_file(self, filename="test.rtf"):
        """
        Write the PRM file
        """

        f = open(filename, "w")
        f.writelines(self.to_string())
        f.close()

    def parse_mass(self, mass):
        self.labels = []
        self.mass = []
        for a in mass:
            tmp = a.split()
            self.labels.append(tmp[2])
            self.mass.append(float(tmp[3]))

    def merge(self, rtf1=None, single=None):
        """
        merge two RTFs into one
        """

        if rtf1 is not None:
            for l, m in zip(rtf1.labels, rtf1.mass):
                if l not in self.labels:
                    self.labels.append(l)
                    self.mass.append(m)

            self.residues += rtf1.residues
            for i, res in enumerate(self.residues):
                residue = {
                    "NAME": str(i) + res["NAME"][:2],
                    "CHARGE": res["CHARGE"],
                    "ATOM": [],
                    "BOND": [],
                    "ANGL": [],
                    "DIHE": [],
                    "IMPH": [],
                }
                for a in res["ATOM"]:
                    tmp = a.split()
                    a1 = str(i) + tmp[1]
                    a = f"ATOM {a1:6s} {tmp[2]:2s}{float(tmp[3]):12.6f}"
                    residue["ATOM"].append(a)
                for a in res["BOND"]:
                    tmp = a.split("!")
                    tmp1 = tmp[0].split()
                    a1, a2 = str(i) + tmp1[1], str(i) + tmp1[2]
                    a = f"BOND {a1:6s} {a2:6s}"
                    if len(tmp) > 1:
                        a += f"!{tmp[-1]:12s}"
                    residue["BOND"].append(a)
                for a in res["ANGL"]:
                    tmp = a.split("!")
                    tmp1 = tmp[0].split()
                    a1, a2, a3 = str(i) + tmp1[1], str(i) + tmp1[2], str(i) + tmp1[3]
                    a = f"ANGL {a1:6s} {a2:6s} {a3:6s} "
                    if len(tmp) > 1:
                        a += f"!{tmp[-1]:12s}"
                    residue["ANGL"].append(a)
                for a in res["DIHE"]:
                    tmp = a.split("!")
                    tmp1 = tmp[0].split()
                    a1, a2, a3, a4 = (
                        str(i) + tmp1[1],
                        str(i) + tmp1[2],
                        str(i) + tmp1[3],
                        str(i) + tmp1[4],
                    )
                    a = f"DIHE {a1:6s} {a2:6s} {a3:6s} {a4:6s} "
                    if len(tmp) > 1:
                        a += f"!{tmp[-1]:12s}"
                    residue["DIHE"].append(a)
                for a in res["IMPH"]:
                    tmp = a.split("!")
                    tmp1 = tmp[0].split()
                    a1, a2, a3, a4 = (
                        str(i) + tmp1[1],
                        str(i) + tmp1[2],
                        str(i) + tmp1[3],
                        str(i) + tmp1[4],
                    )
                    a = f"IMPH {a1:6s} {a2:6s} {a3:6s} {a4:6s}"
                    if len(tmp) > 1:
                        a += f"!{tmp[-1]:12s}"
                    residue["IMPH"].append(a)
                self.residues[i] = residue

        elif single is not None:
            count = str(len(self.residues))
            l = single["label"]
            self.labels.append(l)
            self.mass.append(single["mass"])
            self.residues += [
                {
                    "NAME": single["name"],
                    "CHARGE": single["charge"],
                    "ATOM": ["ATOM {:6s} {:2s}{:12.6f}".format(count + l, l, float(single["charge"]))],
                    "BOND": [],
                    "ANGL": [],
                    "DIHE": [],
                    "IMPH": [],
                }
            ]


class PRM:
    """
    Parse, convert and merge PRM files for CHARMM
    """

    def __init__(self, input):
        self.keywords = ["BOND", "ANGLE", "DIHEDRAL", "IMPHI", "NONBOND"]
        if type(input) == str:
            print("Converting PRM from the file", input)
            f = open(input)
            prm = f.read().split("\n")
            f.close()
        else:
            prm = input

        self.dict = self.parse(prm)

    def parse(self, prm):
        """
        parse prm list to get the dictionary
        """
        dicts = {}
        for key in self.keywords:
            dicts[key] = []

        do_bond = False
        do_angle = False
        do_dihedral = False
        do_imphi = False
        do_nonbond = False

        for l in prm:
            tmp = l.split()
            if l.find("ANGLE") == 0:
                do_angle = True
            elif l.find("BOND") == 0:
                do_bond = True
            elif l.find("DIHEDRAL") == 0:
                do_dihedral = True
            elif l.find("IMPHI") == 0:
                do_imphi = True
            elif l.find("NONBOND") == 0:
                do_nonbond = True

            elif do_bond:
                if len(tmp) > 0:
                    dicts["BOND"].append(l)
                else:
                    do_bond = False

            elif do_angle:
                if len(tmp) > 0:
                    dicts["ANGLE"].append(l)
                else:
                    do_angle = False

            elif do_dihedral:
                if len(tmp) > 0:
                    dicts["DIHEDRAL"].append(l)
                else:
                    do_dihedral = False

            elif do_imphi:
                if len(tmp) > 0:
                    dicts["IMPHI"].append(l)
                else:
                    do_imphi = False

            elif do_nonbond:
                if len(tmp) > 0:
                    if len(tmp) == 7 and tmp[0] != "!":
                        dicts["NONBOND"].append(l)
                else:
                    do_nonbond = False

        return dicts

    def to_string(self):
        """
        convert the PRM object to string
        """

        strs = "* Force Field Parameter File.\n"
        strs += "*"
        for key in self.keywords:
            if key == "NONBOND":
                strs += "\n\nNONBONDED  NBXMOD 5  GROUP SWITCH CDIEL - \n"
                strs += "CUTNB 14.0  CTOFNB 12.0  CTONNB 10.0  EPS 1.0  E14FAC 0.83333333  WMIN 1.4\n"
                strs += "!                Emin     Rmin/2              Emin/2     Rmin  (for 1-4's)\n"
                strs += "!             (kcal/mol)    (A)\n"
            else:
                strs += f"\n\n{key:s}\n"
            for l in self.dict[key]:
                strs += f"{l:s}\n"

        return strs

    def to_file(self, filename="test.prm"):
        """
        Write the PRM file
        """

        f = open(filename, "w")
        f.writelines(self.to_string())
        f.close()

    def merge(self, prm1=None, single=None):
        """
        merge PRM into one
        """
        if prm1 is not None:
            for key in self.keywords:
                for data in prm1.dict[key]:
                    if data not in self.dict[key]:
                        self.dict[key].append(data)
        elif single is not None:
            self.dict["NONBOND"] += [single["nonbond"]]
            # add the nonbonded parameters


if __name__ == "__main__":
    from pyxtal.db import database

    w_dir = "tmp"
    if not os.path.exists(w_dir):
        os.makedirs(w_dir)

    db = database("benchmarks/test.db")
    row = db.get_row("ACSALA")
    struc = db.get_pyxtal("ACSALA")
    c_info = row.data["charmm_info"]
    prm = open(w_dir + "/pyxtal.prm", "w")
    prm.write(c_info["prm"])
    prm.close()
    rtf = open(w_dir + "/pyxtal.rtf", "w")
    rtf.write(c_info["rtf"])
    rtf.close()

    calc = CHARMM(struc, atom_info=c_info, folder=w_dir)
    print(calc.structure.lattice)
    calc.run(clean=False)
    print(calc.structure.energy)
    print(calc.structure.lattice)
